@@ -10,12 +10,21 @@
 #include "kanji.h"
 
 #define	MAXTIMESTR	8
+#define	ATTRWIDTH	10
 
 typedef struct _attrib_t {
+	short nlink;
 	u_int mode;
 #ifdef	HAVEFLAGS
 	u_long flags;
 #endif
+#ifndef	_NOEXTRAATTR
+	u_int mask;
+# ifndef	NOUID
+	uid_t uid;
+	gid_t gid;
+# endif
+#endif	/* !_NOEXTRAATTR */
 	char timestr[2][MAXTIMESTR + 1];
 } attrib_t;
 
@@ -24,6 +33,7 @@ extern int subwindow;
 extern int win_x;
 extern int win_y;
 extern int lcmdline;
+extern int maxcmdline;
 extern int mark;
 #ifdef	HAVEFLAGS
 extern u_long fflaglist[];
@@ -45,7 +55,7 @@ extern int inheritcopy;
 #endif
 
 static int NEAR issamedir __P_((char *, char *));
-static int NEAR islowerdir __P_((char *, char *));
+static int NEAR islowerdir __P_((VOID_A));
 static char *NEAR getdestdir __P_((char *, char *));
 static int NEAR getdestpath __P_((char *, char *, struct stat *));
 static int NEAR checkdupl __P_((char *, char *, struct stat *, struct stat *));
@@ -57,7 +67,12 @@ static int touchdir __P_((char *));
 #ifndef	_NOEXTRACOPY
 static int mvdir __P_((char *));
 #endif
-static VOID NEAR showattr __P_((namelist *, attrib_t *, int y));
+static VOID NEAR showmode __P_((attrib_t *, int, int));
+static VOID NEAR showattr __P_((namelist *, attrib_t *, int));
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+static int NEAR inputuid __P_((attrib_t *, int));
+static int NEAR inputgid __P_((attrib_t *, int));
+#endif
 static char **NEAR getdirtree __P_((char *, char **, int *, int));
 static int NEAR _applydir __P_((char *, int (*)(char *),
 		int (*)(char *), int (*)(char *), int, char *, int));
@@ -69,12 +84,21 @@ int copypolicy = 0;
 int removepolicy = 0;
 char *destpath = NULL;
 
+static short attrnlink = 0;
 static u_short attrmode = 0;
 #ifdef	HAVEFLAGS
 static u_long attrflags = 0;
 #endif
+#ifndef	_NOEXTRAATTR
+static u_short attrmask = 0;
+# ifndef	NOUID
+static uid_t attruid = (uid_t)-1;
+static gid_t attrgid = (gid_t)-1;
+# endif
+#endif	/* !_NOEXTRAATTR */
 static time_t attrtime = 0;
 static char *destdir = NULL;
+static short destnlink = 0;
 static time_t destmtime = (time_t)-1;
 static time_t destatime = (time_t)-1;
 #ifndef	_NOEXTRACOPY
@@ -119,16 +143,17 @@ char *path, *org;
 	return(i);
 }
 
-static int NEAR islowerdir(path, org)
-char *path, *org;
+static int NEAR islowerdir(VOID_A)
 {
 #ifdef	_USEDOSEMU
 	char orgpath[MAXPATHLEN];
 #endif
-	char *cp, *top, *cwd, tmp[MAXPATHLEN];
+	char *cp, *top, *cwd, *path, *org, tmp[MAXPATHLEN];
 	int i;
 
 	cwd = getwd2();
+	path = destpath;
+	org = filelist[filepos].name;
 	if (!org) org = cwd;
 #ifdef	_USEDOSEMU
 	else org = nodospath(orgpath, org);
@@ -177,7 +202,7 @@ char *mes, *arg;
 #endif
 
 	if (arg && *arg) dir = strdup2(arg);
-	else if (!(dir = inputstr(mes, 1, -1, NULL, 1))) return(NULL);
+	else if (!(dir = inputstr(mes, 1, -1, NULL, HST_PATH))) return(NULL);
 	else if (!*(dir = evalpath(dir, 0))) {
 		dir = realloc2(dir, 2);
 		dir[0] = '.';
@@ -413,8 +438,8 @@ int mode;
 		}
 		if (s_islnk(&st)) return(0);
 
-		if (!(tmp = strrdelim(path, 0))) strcpy(dir, ".");
-		else if (tmp == path) strcpy(dir, _SS_);
+		if (!(tmp = strrdelim(path, 0))) copycurpath(dir);
+		else if (tmp == path) copyrootpath(dir);
 		else strncpy2(dir, path, tmp - path);
 
 		if (nodoslstat(dir, &st) < 0) {
@@ -506,6 +531,7 @@ char *path;
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
 
+	destnlink = (TCH_MODE | TCH_UID | TCH_GID);
 	switch (checkdupl(path, dest, &st1, &st2)) {
 		case 2:
 		/* Already exist, but not directory */
@@ -519,6 +545,7 @@ char *path;
 			if (Xmkdir(dest, st1.st_mode & 0777) < 0)
 #endif
 				return(-1);
+			destnlink |= (TCH_ATIME | TCH_MTIME);
 			destmtime = st1.st_mtime;
 			destatime = st1.st_atime;
 			break;
@@ -530,6 +557,7 @@ char *path;
 			break;
 		default:
 		/* Already exist */
+			destnlink |= (TCH_ATIME | TCH_MTIME);
 			destmtime = st2.st_mtime;
 			destatime = st2.st_atime;
 			break;
@@ -555,13 +583,9 @@ char *path;
 	if (!inheritcopy) return(0);
 # endif
 	if (getdestpath(path, dest, &st) < 0) return(-2);
-	if (destmtime != (time_t)-1 || destatime != (time_t)-1) {
-		st.st_mtime = destmtime;
-		st.st_atime = destatime;
-	}
-# ifdef	HAVEFLAGS
-	st.st_flags = (u_long)-1;
-# endif
+	st.st_nlink = destnlink;
+	st.st_mtime = destmtime;
+	st.st_atime = destatime;
 	if (touchfile(dest, &st) < 0) return(-1);
 #endif	/* _USEDOSCOPY || !_NOEXTRACOPY */
 
@@ -577,13 +601,9 @@ char *path;
 	int n;
 
 	if (getdestpath(path, dest, &st) < 0) return(-2);
-	if (destmtime != (time_t)-1 || destatime != (time_t)-1) {
-		st.st_mtime = destmtime;
-		st.st_atime = destatime;
-	}
-# ifdef	HAVEFLAGS
-	st.st_flags = (u_long)-1;
-# endif
+	st.st_nlink = destnlink;
+	st.st_mtime = destmtime;
+	st.st_atime = destatime;
 	if (touchfile(dest, &st) < 0) return(-1);
 	if ((n = checkrmv(path, R_OK | W_OK | X_OK)) < 0)
 		return((n < -1) ? -2 : 1);
@@ -643,14 +663,47 @@ char *path;
 	return(0);
 }
 
-static VOID NEAR showattr(namep, attr, y)
+static VOID NEAR showmode(attr, x, y)
+attrib_t *attr;
+int x, y;
+{
+#ifndef	_NOEXTRAATTR
+# if	!MSDOS
+	u_int tmp;
+# endif
+	char mask[WMODE + 1];
+	int i;
+#endif	/* !_NOEXTRAATTR */
+	char buf[WMODE + 1];
+
+	Xlocate(x, y);
+	putmode(buf, attr -> mode, 1);
+#ifndef	_NOEXTRAATTR
+	putmode(mask, attr -> mask, 1);
+	for (i = 0; buf[i] && mask[i]; i++) {
+		if (mask[i] != '-') buf[i] = '*';
+# if	!MSDOS
+		else if (!((i + 1) % 3)) {
+			tmp = (1 << (12 + 3 - ((i + 1) / 3)));
+			if (attr -> mode & tmp) {
+				if (buf[i] == '-') buf[i] = '!';
+				else buf[i] = 'X';
+			}
+		}
+# endif
+	}
+#endif	/* !_NOEXTRAATTR */
+	Xcputs2(buf);
+}
+
+static VOID NEAR showattr(namep, attr, yy)
 namelist *namep;
 attrib_t *attr;
-int y;
+int yy;
 {
 	struct tm *tm;
 	char buf[WMODE + 1];
-	int x1, x2, w;
+	int x1, x2, y, w;
 
 	tm = localtime(&(namep -> st_mtim));
 	if (isbestomit()) {
@@ -662,7 +715,14 @@ int y;
 		x1 = n_column / 2 - 20;
 		x2 = n_column / 2;
 		w = 16;
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+		if (iswellomit()) {
+			x1 -= WOWNER / 2;
+			x2 -= WOWNER / 2;
+		}
+#endif
 	}
+	y = yy;
 
 	Xlocate(0, y);
 	Xputterm(L_CLEAR);
@@ -683,9 +743,7 @@ int y;
 	Xlocate(x2, y);
 	putmode(buf, namep -> st_mode, 1);
 	Xcputs2(buf);
-	Xlocate(x2 + 10, y);
-	putmode(buf, attr -> mode, 1);
-	Xcputs2(buf);
+	showmode(attr, x2 + ATTRWIDTH, y);
 
 #ifdef	HAVEFLAGS
 	Xlocate(0, ++y);
@@ -695,7 +753,7 @@ int y;
 	Xlocate(x2, y);
 	putflags(buf, namep -> st_flags);
 	Xcputs2(buf);
-	Xlocate(x2 + 10, y);
+	Xlocate(x2 + ATTRWIDTH, y);
 	putflags(buf, attr -> flags);
 	Xcputs2(buf);
 #endif
@@ -707,7 +765,7 @@ int y;
 	Xlocate(x2, y);
 	Xcprintf2("%02d-%02d-%02d",
 		tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday);
-	Xlocate(x2 + 10, y);
+	Xlocate(x2 + ATTRWIDTH, y);
 	Xcputs2(attr -> timestr[0]);
 
 	Xlocate(0, ++y);
@@ -717,22 +775,129 @@ int y;
 	Xlocate(x2, y);
 	Xcprintf2("%02d:%02d:%02d",
 		tm -> tm_hour, tm -> tm_min, tm -> tm_sec);
-	Xlocate(x2 + 10, y);
+	Xlocate(x2 + ATTRWIDTH, y);
 	Xcputs2(attr -> timestr[1]);
 
 	Xlocate(0, ++y);
 	Xputterm(L_CLEAR);
+
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+	if (ishardomit()) return;
+
+	y = yy + 1;
+	x2 += 20;
+	Xlocate(x2, y++);
+	Xkanjiputs(TOWN_K);
+	Xlocate(x2, y++);
+	putowner(buf, attr -> uid);
+	Xputch2('<');
+	if (attr -> nlink & TCH_UID) Xputterm(T_STANDOUT);
+	Xkanjiputs(buf);
+	if (attr -> nlink & TCH_UID) Xputterm(END_STANDOUT);
+	Xputch2('>');
+# ifdef	HAVEFLAGS
+	y++;
+# endif
+	Xlocate(x2, y++);
+	Xkanjiputs(TGRP_K);
+	Xlocate(x2, y++);
+	putgroup(buf, attr -> gid);
+	Xputch2('<');
+	if (attr -> nlink & TCH_GID) Xputterm(T_STANDOUT);
+	Xkanjiputs(buf);
+	if (attr -> nlink & TCH_GID) Xputterm(END_STANDOUT);
+	Xputch2('>');
+#endif	/* !_NOEXTRAATTR && !NOUID */
 }
 
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+static int NEAR inputuid(attr, yy)
+attrib_t *attr;
+int yy;
+{
+	uidtable *up;
+	char *cp, *s, buf[MAXLONGWIDTH + 1];
+	uid_t uid;
+
+	up = finduid(attr -> uid, NULL);
+	if (up) cp = up -> name;
+	else {
+		snprintf2(buf, sizeof(buf), "%-d", (int)(attr -> uid));
+		cp = buf;
+	}
+
+	yy += 2 + WMODELINE + 2;
+	lcmdline = yy;
+	maxcmdline = 1;
+	if (!(s = inputstr(AOWNR_K, 0, -1, cp, HST_USER))) return(-1);
+	if ((cp = sscanf2(s, "%-*d%$", sizeof(uid_t), &uid))) /*EMPTY*/;
+	else if ((up = finduid(0, s))) uid = up -> uid;
+	else {
+		lcmdline = yy;
+		warning(ENOENT, s);
+		free(s);
+		return(-1);
+	}
+
+	free(s);
+	if (uid != attr -> uid) {
+		attr -> uid = uid;
+		attr -> nlink |= TCH_UID;
+	}
+	return(0);
+}
+
+static int NEAR inputgid(attr, yy)
+attrib_t *attr;
+int yy;
+{
+	gidtable *gp;
+	char *cp, *s, buf[MAXLONGWIDTH + 1];
+	gid_t gid;
+
+	gp = findgid(attr -> gid, NULL);
+	if (gp) cp = gp -> name;
+	else {
+		snprintf2(buf, sizeof(buf), "%-d", (int)(attr -> gid));
+		cp = buf;
+	}
+
+	yy += 2 + WMODELINE + 2;
+	lcmdline = yy;
+	maxcmdline = 1;
+	if (!(s = inputstr(AGRUP_K, 0, -1, cp, HST_GROUP))) return(-1);
+	if ((cp = sscanf2(s, "%-*d%$", sizeof(gid_t), &gid))) /*EMPTY*/;
+	else if ((gp = findgid(0, s))) gid = gp -> gid;
+	else {
+		lcmdline = yy;
+		warning(ENOENT, s);
+		free(s);
+		return(-1);
+	}
+
+	free(s);
+	if (gid != attr -> gid) {
+		attr -> gid = gid;
+		attr -> nlink |= TCH_GID;
+	}
+	return(0);
+}
+#endif	/* !_NOEXTRAATTR && !NOUID */
+
 int inputattr(namep, flag)
 namelist *namep;
 int flag;
 {
+#if	!MSDOS
+	u_int tmp;
+#endif
+#ifdef	HAVEFLAGS
+	char buf[WMODE + 1];
+#endif
 	struct tm *tm;
 	attrib_t attr;
-	char buf[WMODE + 1];
-	u_int tmp;
-	int i, ch, x, y, xx, yy, ymin, ymax, dupwin_x, dupwin_y;
+	u_int mask;
+	int ch, x, y, xx, yy, ymin, ymax, dupwin_x, dupwin_y, excl;
 
 	dupwin_x = win_x;
 	dupwin_y = win_y;
@@ -742,9 +907,26 @@ int flag;
 	yy = filetop(win);
 	while (yy + WMODELINE + 5 > n_line - 1) yy--;
 	if (yy <= L_TITLE) yy = L_TITLE + 1;
-	xx = n_column / 2 + ((isbestomit()) ? 7 : 10);
+	xx = n_column / 2 + ((isbestomit()) ? 7 : ATTRWIDTH);
 
+	excl = (flag & ATR_EXCLUSIVE);
+	attr.nlink = TCH_CHANGE;
+	if (flag & ATR_MULTIPLE) attr.nlink |= (TCH_MODE | TCH_MTIME);
 	attr.mode = namep -> st_mode;
+#ifndef	_NOEXTRAATTR
+	attr.mode &= ~S_IFMT;
+	attr.mask = 0;
+# if	!MSDOS
+	if ((flag & ATR_RECURSIVE) && excl == ATR_MODEONLY) {
+		for (x = 0; x < 3; x++) attr.mode |= (1 << (x + 12));
+		attr.nlink |= TCH_MODEEXE;
+	}
+# endif
+# ifndef	NOUID
+	attr.uid = namep -> st_uid;
+	attr.gid = namep -> st_gid;
+# endif
+#endif	/* !_NOEXTRAATTR */
 #ifdef	HAVEFLAGS
 	attr.flags = namep -> st_flags;
 #endif
@@ -754,8 +936,12 @@ int flag;
 	snprintf2(attr.timestr[1], sizeof(attr.timestr[1]), "%02d:%02d:%02d",
 		tm -> tm_hour, tm -> tm_min, tm -> tm_sec);
 	showattr(namep, &attr, yy);
-	y = ymin = (flag & 1) ? 0 : WMODELINE;
-	ymax = (flag & 2) ? WMODELINE + 1 : 0;
+	y = ymin = (excl == ATR_TIMEONLY) ? WMODELINE : 0;
+	ymax = (excl == ATR_MODEONLY) ? 0 : WMODELINE + 1;
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+	if (excl == ATR_OWNERONLY) x = ATTRWIDTH + 1;
+	else
+#endif
 	x = 0;
 
 	do {
@@ -765,14 +951,33 @@ int flag;
 		Xtflush();
 
 		keyflush();
+#if	MSDOS
+		if (x == 3) mask = S_ISVTX;
+		else
+#endif
+		mask = (1 << (8 - x));
 		switch (ch = Xgetkey(1, 0)) {
 			case K_UP:
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+				if (x > ATTRWIDTH) {
+					if (y) y = 0;
+					else y = WMODELINE + 1;
+					break;
+				}
+#endif
 				if (y > ymin) y--;
 				else y = ymax;
 				if (y && x >= 8) x = 7;
 				if (!y && x >= WMODE - 1) x = WMODE - 2;
 				break;
 			case K_DOWN:
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+				if (x > ATTRWIDTH) {
+					if (y) y = 0;
+					else y = WMODELINE + 1;
+					break;
+				}
+#endif
 				if (y < ymax) y++;
 				else y = ymin;
 				if (y && x >= 8) x = 7;
@@ -780,22 +985,43 @@ int flag;
 				break;
 			case 'a':
 			case 'A':
-				if (ymin <= 0) x = y = 0;
+				if (excl && excl != ATR_MODEONLY) break;
+				x = y = 0;
 				break;
 			case 'd':
 			case 'D':
-				if (ymax >= WMODELINE) {
-					x = 0;
-					y = WMODELINE;
-				}
+				if (excl && excl != ATR_TIMEONLY) break;
+				x = 0;
+				y = WMODELINE;
 				break;
 			case 't':
 			case 'T':
-				if (ymax >= WMODELINE + 1) {
-					x = 0;
-					y = WMODELINE + 1;
-				}
+				if (excl && excl != ATR_TIMEONLY) break;
+				x = 0;
+				y = WMODELINE + 1;
 				break;
+#ifdef	HAVEFLAGS
+			case 'f':
+			case 'F':
+				if (excl) break;
+				x = 0;
+				y = WMODELINE - 1;
+				break;
+#endif
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+			case 'o':
+			case 'O':
+				if (excl && excl != ATR_OWNERONLY) break;
+				x = ATTRWIDTH + 1;
+				y = 0;
+				break;
+			case 'g':
+			case 'G':
+				if (excl && excl != ATR_OWNERONLY) break;
+				x = ATTRWIDTH + 1;
+				y = WMODELINE + 1;
+				break;
+#endif	/* !_NOEXTRAATTR && !NOUID */
 			case '0':
 			case '1':
 			case '2':
@@ -809,6 +1035,7 @@ int flag;
 				if (y < WMODELINE) break;
 				Xputch2(ch);
 				attr.timestr[y - WMODELINE][x] = ch;
+				attr.nlink |= TCH_MTIME;
 /*FALLTHRU*/
 			case K_RIGHT:
 #ifdef	HAVEFLAGS
@@ -819,6 +1046,13 @@ int flag;
 #endif
 				if (y) {
 					if (x >= 7) {
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+						if (!excl
+						&& y == WMODELINE + 1) {
+							x = ATTRWIDTH + 1;
+							break;
+						}
+#endif
 						if (y == WMODELINE + 1) break;
 						y++;
 						x = 0;
@@ -829,11 +1063,22 @@ int flag;
 					}
 				}
 				else if (x < WMODE - 2) x++;
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+				else if (!excl) x = ATTRWIDTH + 1;
+#endif
 				break;
 			case K_BS:
 				if (y < WMODELINE) break;
 /*FALLTHRU*/
 			case K_LEFT:
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+				if (x > ATTRWIDTH) {
+					if (excl || ch == K_BS) break;
+					else if (y) x = 7;
+					else x = WMODE - 2;
+				}
+				else
+#endif
 #ifdef	HAVEFLAGS
 				if (y == WMODELINE - 1) {
 					if (x > 0) x--;
@@ -857,43 +1102,84 @@ int flag;
 				yy = filetop(win);
 				while (yy + WMODELINE + 5 > n_line - 1) yy--;
 				if (yy <= L_TITLE) yy = L_TITLE + 1;
-				xx = n_column / 2 + ((isbestomit()) ? 7 : 10);
+				xx = n_column / 2
+					+ ((isbestomit()) ? 7 : ATTRWIDTH);
 				showattr(namep, &attr, yy);
 				break;
 			case ' ':
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+				if (x > ATTRWIDTH) {
+					if (y) inputgid(&attr, yy);
+					else inputuid(&attr, yy);
+					showattr(namep, &attr, yy);
+					break;
+				}
+#endif
 #ifdef	HAVEFLAGS
 				if (y == WMODELINE - 1) {
 					attr.flags ^= fflaglist[x];
 					Xlocate(xx, yy + y + 2);
 					putflags(buf, attr.flags);
 					Xcputs2(buf);
+					attr.nlink |= TCH_FLAGS;
 					break;
 				}
 #endif
 				if (y) break;
 #if	MSDOS
 				if (x == 2) break;
-				else if (x == 3) attr.mode ^= S_ISVTX;
+#else	/* !MSDOS */
+				if (!((x + 1) % 3) && (attr.mode & mask)) {
+					tmp = (1 << (12 - ((x + 1) / 3)));
+# ifndef	_NOEXTRAATTR
+					if (flag & ATR_RECURSIVE) tmp <<= 3;
+# endif
+					if (attr.mode & tmp) {
+# ifndef	_NOEXTRAATTR
+						if (flag & ATR_RECURSIVE)
+							/*EMPTY*/;
+						else
+# endif
+						attr.mode ^= tmp;
+					}
+# ifndef	_NOEXTRAATTR
+					else if (flag & ATR_RECURSIVE) {
+						mask = (tmp >> 3);
+						if (attr.mode & mask)
+							attr.mode ^= tmp;
+					}
+# endif
+					else mask = tmp;
+				}
+# ifndef	_NOEXTRAATTR
 				else {
-					tmp = 1;
-					for (i = 8; i > x; i--) tmp <<= 1;
-					attr.mode ^= tmp;
+					tmp = (1 << (15 - ((x + 1) / 3)));
+					if (attr.mode & tmp) mask = tmp;
 				}
-#else
-				tmp = 1;
-				for (i = 8; i > x; i--) tmp <<= 1;
-				if (!((x + 1) % 3) && (attr.mode & tmp)) {
-					i = (x * 2) / 3 + 4;
-					if (!(attr.mode & (tmp << i)))
-						tmp <<= i;
-					else attr.mode ^= (tmp << i);
-				}
-				attr.mode ^= tmp;
-#endif
-				Xlocate(xx, yy + y + 2);
-				putmode(buf, attr.mode, 1);
-				Xcputs2(buf);
+# endif
+#endif	/* !MSDOS */
+				attr.mode ^= mask;
+				showmode(&attr, xx, yy + y + 2);
+				attr.nlink |= TCH_MODE;
 				break;
+#ifndef	_NOEXTRAATTR
+			case 'm':
+			case 'M':
+# ifndef	NOUID
+				if (x > ATTRWIDTH) break;
+# endif
+				if (y || !(flag & ATR_MULTIPLE)) break;
+# if	MSDOS
+				if (x == 2) break;
+# else
+				if (!((x + 1) % 3))
+					mask |= (1 << (12 - ((x + 1) / 3)));
+# endif
+				attr.mask ^= mask;
+				showmode(&attr, xx, yy + y + 2);
+				attr.nlink |= TCH_MASK;
+				break;
+#endif	/* !_NOEXTRAATTR */
 			default:
 				break;
 		}
@@ -926,25 +1212,44 @@ int flag;
 	|| tm -> tm_hour > 23 || tm -> tm_min > 59 || tm -> tm_sec > 59)
 		return(-1);
 
+	mask = (TCH_CHANGE | TCH_MASK | TCH_MODEEXE);
+	switch (excl) {
+		case ATR_MODEONLY:
+			mask |= TCH_MODE;
+			break;
+		case ATR_TIMEONLY:
+			mask |= TCH_MTIME;
+			break;
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+		case ATR_OWNERONLY:
+			mask |= (TCH_UID | TCH_GID);
+			break;
+#endif
+		default:
+			if (attrmode != namep -> st_mode) mask |= TCH_MODE;
+#ifdef	HAVEFLAGS
+			if (attrflags != namep -> st_flags) mask |= TCH_FLAGS;
+#endif
+#if	!defined (_NOEXTRAATTR) && !defined (NOUID)
+			if (attruid != namep -> st_uid) mask |= TCH_UID;
+			if (attrgid != namep -> st_gid) mask |= TCH_GID;
+#endif
+			if (attrtime != namep -> st_mtim) mask |= TCH_MTIME;
+			break;
+	}
+	attrnlink = (attr.nlink & mask);
 	attrmode = attr.mode;
 #ifdef	HAVEFLAGS
 	attrflags = attr.flags;
 #endif
+#ifndef	_NOEXTRAATTR
+	attrmask = attr.mask;
+# ifndef	NOUID
+	attruid = attr.uid;
+	attrgid = attr.gid;
+# endif
+#endif	/* !_NOEXTRAATTR */
 	attrtime = timelocal2(tm);
-	if (flag == 3) {
-		if (attrmode == namep -> st_mode) attrmode = (u_short)-1;
-#ifdef	HAVEFLAGS
-		if (attrflags == namep -> st_flags) attrflags = (u_long)-1;
-#endif
-		if (attrtime == namep -> st_mtim) attrtime = (time_t)-1;
-	}
-	else {
-		if (!(flag & 1)) attrmode = (u_short)-1;
-#ifdef	HAVEFLAGS
-		attrflags = (u_long)-1;
-#endif
-		if (!(flag & 2)) attrtime = (time_t)-1;
-	}
 
 	return(1);
 }
@@ -954,14 +1259,19 @@ char *path;
 {
 	struct stat st;
 
-	st.st_mtime = attrtime;
+	st.st_nlink = attrnlink;
+	st.st_mode = attrmode;
 #ifdef	HAVEFLAGS
 	st.st_flags = attrflags;
 #endif
-	st.st_mode = attrmode;
-#ifndef	NOUID
-	st.st_gid = (gid_t)-1;
-#endif
+#ifndef	_NOEXTRAATTR
+	st.st_size = (off_t)attrmask;
+# ifndef	NOUID
+	st.st_uid = attruid;
+	st.st_gid = attrgid;
+# endif
+#endif	/* !_NOEXTRAATTR */
+	st.st_mtime = attrtime;
 
 	return(touchfile(path, &st));
 }
@@ -1060,7 +1370,7 @@ int verbose;
 	struct stat st;
 	time_t dupmtime, dupatime;
 	char *cp, *fname, path[MAXPATHLEN], **dirlist;
-	int ret, ndir, max;
+	int ret, ndir, max, dupnlink;
 
 	if (intrkey()) return(-2);
 
@@ -1075,16 +1385,18 @@ int verbose;
 	else cp = NULL;	/* fake for -Wuninitialized */
 #endif
 
-	if (!funcd1 && !funcd2) order = 0;
+	if (!funcd1) order = (funcd2) ? ORD_NOPREDIR : ORD_NODIR;
 	strcpy(path, dir);
 
 	ret = max = 0;
 	if (!order) dirlist = NULL;
-	else dirlist = getdirtree(path, NULL, &max, (order == 2) ? 0 : 1);
+	else dirlist = getdirtree(path,
+		NULL, &max, (order == ORD_LOWER) ? 0 : 1);
 	fname = strcatdelim(path);
 
-	destmtime = destatime = (time_t)-1;
-	if ((order == 1 || order == 2) && (ret = (*funcd1)(dir)) < 0) {
+	destnlink = 0;
+	if ((order == ORD_NORMAL || order == ORD_LOWER)
+	&& (ret = (*funcd1)(dir)) < 0) {
 		if (ret == -1) warning(-1, dir);
 		if (dirlist) {
 			for (ndir = 0; ndir < max; ndir++) free(dirlist[ndir]);
@@ -1092,11 +1404,12 @@ int verbose;
 		}
 		return(ret);
 	}
+	dupnlink = destnlink;
 	dupmtime = destmtime;
 	dupatime = destatime;
 
-	if (order) for (ndir = 0; ndir < max; ndir++) {
-		if (order != 2)
+	if (order != ORD_NODIR) for (ndir = 0; ndir < max; ndir++) {
+		if (order != ORD_LOWER)
 			ret = _applydir(dirlist[ndir], funcf,
 				funcd1, funcd2, order, NULL, verbose);
 		else if ((ret = (*funcd1)(dirlist[ndir])) < 0) {
@@ -1104,7 +1417,7 @@ int verbose;
 			ret = -2;
 		}
 		else ret = _applydir(dirlist[ndir], funcf,
-			funcd1, funcd2, 0, NULL, verbose);
+			funcd1, funcd2, ORD_NODIR, NULL, verbose);
 		if (ret < -1) {
 			for (; ndir < max; ndir++) free(dirlist[ndir]);
 			free(dirlist);
@@ -1142,6 +1455,7 @@ int verbose;
 		if (ret < -1) return(ret);
 	}
 
+	destnlink = dupnlink;
 	destmtime = dupmtime;
 	destatime = dupatime;
 	if (funcd2 && (ret = (*funcd2)(dir)) < 0) {
@@ -1166,10 +1480,10 @@ char *endmes;
 
 	verbose = 1;
 	if (!dir) {
-		dir = ".";
+		dir = curpath;
 		verbose = 0;
 	}
-	else if (dir[0] == '.' && dir[1] == '.' && !dir[2]) {
+	else if (isdotdir(dir) == 1) {
 		realpath2(dir, path, 0);
 		dir = path;
 	}
@@ -1185,9 +1499,11 @@ int copyfile(arg, tr)
 char *arg;
 int tr;
 {
+	int order;
+
 	if (!mark && isdotdir(filelist[filepos].name)) {
 		Xputterm(T_BELL);
-		return(0);
+		return(FNC_NONE);
 	}
 #ifdef	_NOTREE
 	destpath = getdestdir(COPYD_K, arg);
@@ -1199,11 +1515,11 @@ int tr;
 # endif
 #endif	/* !_NOTREE */
 
-	if (!destpath) return((tr) ? 2 : 1);
+	if (!destpath) return((tr) ? FNC_UPDATE : FNC_CANCEL);
 	destdir = NULL;
 	copypolicy = (issamedir(destpath, NULL)) ? (FLAG_SAMEDIR | 2) : 0;
 #ifndef	_NODOSDRIVE
-	if (dospath3("")) waitmes();
+	if (dospath3(nullstr)) waitmes();
 #endif
 	if (mark > 0
 	|| (!isdir(&(filelist[filepos])) || islink(&(filelist[filepos]))))
@@ -1213,12 +1529,12 @@ int tr;
 		if (copypolicy) {
 			warning(EEXIST, filelist[filepos].name);
 			free(destpath);
-			return((tr) ? 2 : 1);
+			return((tr) ? FNC_UPDATE : FNC_CANCEL);
 		}
 #endif
-		applydir(filelist[filepos].name, safecopy, cpdir, touchdir,
-			(islowerdir(destpath, filelist[filepos].name)) ? 2 : 1,
-			ENDCP_K);
+		order = (islowerdir()) ? ORD_LOWER : ORD_NORMAL;
+		applydir(filelist[filepos].name, safecopy,
+			cpdir, touchdir, order, ENDCP_K);
 	}
 
 	free(destpath);
@@ -1235,7 +1551,7 @@ int tr;
 	forwarddrive = -1;
 #endif
 
-	return(4);
+	return(FNC_EFFECT);
 }
 
 /*ARGSUSED*/
@@ -1245,12 +1561,15 @@ int tr;
 {
 #ifdef	_USEDOSEMU
 	char path[MAXPATHLEN];
+#endif
+#ifndef	_NOEXTRACOPY
+	int order;
 #endif
 	int i;
 
 	if (!mark && isdotdir(filelist[filepos].name)) {
 		Xputterm(T_BELL);
-		return(0);
+		return(FNC_NONE);
 	}
 #ifdef	_NOTREE
 	destpath = getdestdir(MOVED_K, arg);
@@ -1262,30 +1581,31 @@ int tr;
 # endif
 #endif	/* !_NOTREE */
 
-	if (!destpath || issamedir(destpath, NULL)) return((tr) ? 2 : 1);
+	if (!destpath || issamedir(destpath, NULL))
+		return((tr) ? FNC_UPDATE : FNC_CANCEL);
 	destdir = NULL;
 	copypolicy = removepolicy = 0;
 	if (mark > 0) filepos = applyfile(safemove, ENDMV_K);
-	else if (islowerdir(destpath, filelist[filepos].name))
-		warning(EINVAL, filelist[filepos].name);
+	else if (islowerdir()) warning(EINVAL, filelist[filepos].name);
 	else {
 		i = safemove(fnodospath(path, filepos));
 		if (!i) filepos++;
 		else if (i == -1) {
 #ifdef	_NOEXTRACOPY
 			warning(-1, filelist[filepos].name);
-#else
+#else	/* !_NOEXTRACOPY */
 			if (errno != EXDEV
 			|| !isdir(&(filelist[filepos]))
 			|| islink(&(filelist[filepos])))
 				warning(-1, filelist[filepos].name);
-			else if (applydir(filelist[filepos].name,
-				safemove, cpdir, mvdir,
-				(islowerdir(destpath, filelist[filepos].name))
-				? 2 : 1,
-				ENDMV_K) >= 0)
-					filepos++;
-#endif
+			else {
+				order = (islowerdir())
+					? ORD_LOWER : ORD_NORMAL;
+				i = applydir(filelist[filepos].name, safemove,
+					cpdir, mvdir, order, ENDMV_K);
+				if (i >= 0) filepos++;
+			}
+#endif	/* !_NOEXTRACOPY */
 		}
 	}
 
@@ -1304,7 +1624,7 @@ int tr;
 	forwarddrive = -1;
 #endif
 
-	return(4);
+	return(FNC_EFFECT);
 }
 
 static int forcecpfile(path)
@@ -1319,6 +1639,7 @@ char *path;
 	st.st_flags = (u_long)-1;
 #endif
 
+	st.st_nlink = (TCH_MODE | TCH_UID | TCH_GID | TCH_ATIME | TCH_MTIME);
 	return(touchfile(dest, &st));
 }
 
@@ -1359,6 +1680,7 @@ char *path;
 	st.st_flags = (u_long)-1;
 #endif
 
+	st.st_nlink = (TCH_MODE | TCH_UID | TCH_GID | TCH_ATIME | TCH_MTIME);
 	return(touchfile(dest, &st));
 }
 
