@@ -13,10 +13,14 @@
 #include "func.h"
 #include "termemu.h"
 
+#ifndef	_NOORIGSHELL
+#include "system.h"
+#endif
+
 #define	MAXESCPARAM	16
 #define	MAXESCCHAR	4
 #define	MAXTABSTOP	255
-#define	SIZEFMT		"\033[%d;%dR"
+#define	MAXKANJIBUF	(3 + 2 + 3)
 
 typedef struct _ptyterm_t {
 	short cur_x, cur_y;
@@ -49,8 +53,10 @@ typedef struct _ptyterm_t {
 #define	T_NOAPPLIKEY	00004
 #define	T_NOAPPLICURSOR	00010
 #define	T_CODEALT	00020
+#define	T_LOCKED	00040
 
 extern int hideclock;
+extern int fdmode;
 extern int wheader;
 extern int emufd;
 
@@ -79,6 +85,10 @@ static int wintr __P_((VOID_A));
 static VOID NEAR evalsignal __P_((VOID_A));
 #endif
 static int NEAR convkey __P_((int, keyseq_t *));
+#ifndef	_NOKANJICONV
+static int NEAR ptygetch __P_((int));
+static char *NEAR ptykconv __P_((char *, char *, int, int));
+#endif
 static VOID NEAR evalscroll __P_((int, int, int));
 static VOID NEAR evallf __P_((int));
 static VOID NEAR evalnormal __P_((int, int));
@@ -139,7 +149,12 @@ int w, clean;
 
 	pty[w].min_x = (short)0;
 	pty[w].max_x = n_column;
-	if (w == MAXWINDOWS) {
+#ifdef	_NOORIGSHELL
+	if (w == MAXWINDOWS)
+#else
+	if (w == MAXWINDOWS || isshptymode())
+#endif
+	{
 		pty[w].min_y = (short)0;
 		pty[w].max_y = n_line;
 	}
@@ -199,6 +214,7 @@ int w;
 
 	if (last_attr != pty[w].attr) {
 		putterm(T_NORMAL);
+		last_fg = last_bg = (short)-1;
 		if (pty[w].attr & A_BOLD) putterm(T_BOLD);
 		if (pty[w].attr & A_REVERSE) putterm(T_REVERSE);
 		if (pty[w].attr & A_DIM) putterm(T_DIM);
@@ -213,6 +229,7 @@ int w;
 			tputs2(cp, 1);
 			free(cp);
 		}
+		else cprintf2("\033[%dm", pty[w].fg + ANSI_NORMAL);
 		last_fg = pty[w].fg;
 	}
 	if (last_bg != pty[w].bg && pty[w].bg >= (short)0) {
@@ -220,6 +237,7 @@ int w;
 			tputs2(cp, 1);
 			free(cp);
 		}
+		else cprintf2("\033[%dm", pty[w].bg + ANSI_REVERSE);
 		last_bg = pty[w].bg;
 	}
 }
@@ -351,6 +369,32 @@ keyseq_t *kp;
 	return((getdefkeyseq(kp) < 0) ? getkeyseq(kp) : 0);
 }
 
+#ifndef	_NOKANJICONV
+static int NEAR ptygetch(fd)
+int fd;
+{
+	u_short ch;
+
+	if (recvbuf(fd, &ch, sizeof(ch)) < 0 || (ch & 0xff00)) return(-1);
+	return(ch);
+}
+
+static char *NEAR ptykconv(buf, buf2, incode, outcode)
+char *buf, *buf2;
+int incode, outcode;
+{
+	char *cp, *tmp;
+
+	tmp = kanjiconv2(buf2, buf, MAXKANJIBUF, incode, DEFCODE, L_TERMINAL);
+	if (kanjierrno) return(buf);
+	cp = (tmp == buf) ? buf2 : buf;
+	cp = kanjiconv2(cp, tmp, MAXKANJIBUF, DEFCODE, outcode, L_TERMINAL);
+	if (kanjierrno) return(tmp);
+
+	return(cp);
+}
+#endif	/* !_NOKANJICONV */
+
 static VOID NEAR evalscroll(w, n, c)
 int w, n, c;
 {
@@ -375,14 +419,16 @@ static VOID NEAR evalnormal(w, c)
 int w, c;
 {
 #ifndef	_NOKANJICONV
-	int code;
+	char buf[MAXKANJIBUF + 1], buf2[MAXKANJIBUF + 1];
+	int len, code, incode, outcode;
 #endif
 	int width;
 
 #ifndef	_NOKANJICONV
-	code = outputkcode;
-	if (code == NOCNV) code = DEFCODE;
-	if (code == M_UTF8) code = UTF8;
+	outcode = (outputkcode != NOCNV) ? outputkcode : DEFCODE;
+	incode = (w < MAXWINDOWS && ptylist[w].outcode != NOCNV)
+		? ptylist[w].outcode : outcode;
+	code = (incode != M_UTF8) ? incode : UTF8;
 #endif
 
 	if ((pty[w].termflags & T_NOAUTOMARGIN)
@@ -402,9 +448,8 @@ int w, c;
 		}
 		else if (code == UTF8) {
 			if (!ismsb(c)) pty[w].last1 = pty[w].last2 = (short)-1;
-			else if ((pty[w].last1 & 0xc0) == 0x80) /*EMPTY*/;
 			else if (pty[w].last2 >= (short)0) width++;
-			else {
+			else if (!isutf2(pty[w].last1, c)) {
 				pty[w].last2 = c;
 				return;
 			}
@@ -438,9 +483,25 @@ int w, c;
 	surelocate(w, 0);
 	settermattr(w);
 	settermcode(w);
-	if (pty[w].last1 >= (short)0) putch2(pty[w].last1);
-	if (pty[w].last2 >= (short)0) putch2(pty[w].last2);
-	putch2(c);
+#ifndef	_NOKANJICONV
+	if (!(pty[w].termflags & T_MULTIBYTE) && incode != outcode) {
+		len = 0;
+		if (pty[w].last1 >= (short)0) buf[len++] = pty[w].last1;
+		if (pty[w].last2 >= (short)0) buf[len++] = pty[w].last2;
+		if (!len) putch2(c);
+		else {
+			buf[len++] = c;
+			buf[len] = '\0';
+			cputs2(ptykconv(buf, buf2, incode, outcode));
+		}
+	}
+	else
+#endif	/* !_NOKANJICONV */
+	{
+		if (pty[w].last1 >= (short)0) putch2(pty[w].last1);
+		if (pty[w].last2 >= (short)0) putch2(pty[w].last2);
+		putch2(c);
+	}
 	tflush();
 	pty[w].cur_x += width;
 	last_x += width;
@@ -577,7 +638,7 @@ int w, c;
 static VOID NEAR evalcsi(w, c, fd)
 int w, c, fd;
 {
-	char buf[sizeof(SIZEFMT) + 6];
+	char buf[sizeof(SIZEFMT) + (MAXLONGWIDTH - 2) * 2];
 	short min, max;
 	int i, n, x, y;
 
@@ -942,6 +1003,7 @@ int w;
 {
 	u_char uc;
 
+	if (pty[w].termflags & T_LOCKED) return;
 	if (recvbuf(ptylist[w].fd, &uc, sizeof(uc)) < 0) return;
 
 	switch (pty[w].escmode) {
@@ -1071,7 +1133,7 @@ static int NEAR directoutput(fd, n)
 int fd, n;
 {
 	ptyinfo_t tmp;
-	char *s, *arg;
+	char *s, *arg, *cwd;
 	p_id_t pid;
 	short w1, w2, row[MAXWINDOWS];
 	int i;
@@ -1150,7 +1212,6 @@ int fd, n;
 			}
 			else {
 				pty[MAXWINDOWS].fg = w1;
-				pty[MAXWINDOWS].bg = (short)-1;
 			}
 			break;
 		case TE_MOVECURSOR:
@@ -1232,9 +1293,27 @@ int fd, n;
 			return(1);
 /*NOTREACHED*/
 			break;
+		case TE_LOCKBACK:
+			pty[w1].termflags |= T_LOCKED;
+			sendbuf(ptylist[w1].fd, "\n", 1);
+			break;
+		case TE_UNLOCKBACK:
+			pty[w1].termflags &= ~T_LOCKED;
+			sendbuf(ptylist[w1].fd, "\n", 1);
+			break;
 #ifndef	_NOKANJICONV
 		case TE_CHANGEKCODE:
-			outputkcode = w1;
+			if (recvbuf(fd, &w2, sizeof(w2)) < 0) break;
+			inputkcode = w1;
+			outputkcode = w2;
+			break;
+		case TE_CHANGEINKCODE:
+			if (recvbuf(fd, &w2, sizeof(w2)) < 0) break;
+			ptylist[w1].incode = (char)w2;
+			break;
+		case TE_CHANGEOUTKCODE:
+			if (recvbuf(fd, &w2, sizeof(w2)) < 0) break;
+			ptylist[w1].outcode = (char)w2;
 			break;
 #endif	/* !_NOKANJICONV */
 		case TE_AWAKECHILD:
@@ -1245,13 +1324,16 @@ int fd, n;
 				if (s) free(s);
 				break;
 			}
+			if (recvstring(fd, &cwd) < 0) cwd = NULL;
 			resetptyterm(w1, 1);
 
 			sendbuf(ptylist[w1].fd, &n, sizeof(n));
 			sendstring(ptylist[w1].fd, s);
 			sendstring(ptylist[w1].fd, arg);
+			sendstring(ptylist[w1].fd, cwd);
 			if (s) free(s);
 			if (arg) free(arg);
+			if (cwd) free(cwd);
 			break;
 		default:
 			break;
@@ -1263,24 +1345,44 @@ int fd, n;
 static int NEAR evalinput(fd)
 int fd;
 {
+#ifdef	_NOKANJICONV
+	char buf[2];
+#else
+	char buf[MAXKANJIBUF + 1], buf2[MAXKANJIBUF + 1];
+	int cnv, code, incode, outcode;
+#endif
 	keyseq_t key;
-	u_char buf[2];
 	int n;
 
 	if (recvbuf(fd, &(key.code), sizeof((key.code))) < 0) return(0);
 
+#ifndef	_NOKANJICONV
+	cnv = 0;
+	incode = (inputkcode != NOCNV) ? inputkcode : DEFCODE;
+	outcode = (win < MAXWINDOWS && ptylist[win].incode != NOCNV)
+		? ptylist[win].incode : incode;
+	code = (incode != M_UTF8) ? incode : UTF8;
+#endif
+
 	if (ismetakey(key.code)) {
 		key.len = (u_char)2;
-		key.str = (char *)buf;
+		key.str = buf;
 		buf[0] = K_ESC;
 		buf[1] = (key.code & 0xff);
 	}
+#if	!defined (_NOKANJICONV) || defined (CODEEUC)
+# ifdef	_NOKANJICONV
 	else if (isekana2(key.code)) {
+# else
+	else if (inputkcode == EUC && isekana2(key.code)) {
+		if (incode != outcode) cnv++;
+# endif
 		key.len = (u_char)2;
-		key.str = (char *)buf;
+		key.str = buf;
 		buf[0] = C_EKANA;
 		buf[1] = (key.code & 0xff);
 	}
+#endif	/* !_NOKANJICONV || CODEEUC */
 	else if (!(key.code & 01000) && key.code > K_MAX) {
 		n = directoutput(fd, key.code);
 		if (win < MAXWINDOWS) {
@@ -1293,10 +1395,41 @@ int fd;
 	}
 	else if (convkey(win, &key) < 0 || !(key.len)) {
 		key.len = (u_char)1;
-		key.str = (char *)buf;
+		key.str = buf;
 		buf[0] = key.code;
+#ifndef	_NOKANJICONV
+		if (incode == outcode) /*EMPTY*/;
+		else if (inputkcode == SJIS && iskana2(key.code)) cnv++;
+		else if (code == UTF8) {
+			if (!ismsb(key.code)) /*EMPTY*/;
+			else if ((n = ptygetch(fd)) >= 0) {
+				buf[1] = n;
+				(key.len)++;
+				if (isutf2(buf[0], buf[1])) cnv++;
+				else if ((n = ptygetch(fd)) >= 0) {
+					cnv++;
+					buf[2] = n;
+					(key.len)++;
+				}
+			}
+		}
+		else if (isinkanji1(key.code, inputkcode)
+		&& (n = ptygetch(fd)) >= 0) {
+			cnv++;
+			buf[1] = n;
+			(key.len)++;
+		}
+#endif	/* !_NOKANJICONV */
 	}
 
+#ifndef	_NOKANJICONV
+	if (cnv) {
+		buf[key.len] = '\0';
+		key.str = ptykconv(buf, buf2, incode, outcode);
+		key.len = (u_char)strlen(key.str);
+	}
+#endif
+
 	if (win < MAXWINDOWS) sendbuf(ptylist[win].fd, key.str, key.len);
 
 	return(0);
@@ -1305,7 +1438,7 @@ int fd;
 int backend(VOID_A)
 {
 	char result[MAXWINDOWS + 1];
-	int i, n, fds[MAXWINDOWS];
+	int i, n, fds[MAXWINDOWS + 1];
 
 	hideclock = -1;
 	dumbterm = 1;
@@ -1324,11 +1457,12 @@ int backend(VOID_A)
 		evalsignal();
 #endif
 
-		for (i = 0; i < MAXWINDOWS; i++) fds[i] = ptylist[i].fd;
-		if (selectpty(emufd, fds, result, -1) <= 0) continue;
+		for (i = 0; i < MAXWINDOWS; i++)
+			fds[i] = (ptylist[i].pid) ? ptylist[i].fd : -1;
+		fds[i] = emufd;
+		if (selectpty(MAXWINDOWS + 1, fds, result, -1) <= 0) continue;
 
 		if (result[MAXWINDOWS] && (n = evalinput(emufd)) > 0) continue;
-
 		for (i = 0; i < MAXWINDOWS; i++)
 			if (result[i] && ptylist[i].pid) evaloutput(i);
 	}
