@@ -66,6 +66,9 @@ extern int inruncom;
 #ifndef	_NOKANJICONV
 extern int unicodebuffer;
 #endif
+#ifndef	_NOPTY
+extern int parentfd;
+#endif
 
 static VOID NEAR builtinerror __P_((char *[], char *, int));
 #ifdef	_NOORIGSHELL
@@ -116,6 +119,9 @@ static int NEAR kconv __P_((int, char *[]));
 static int NEAR getinputstr __P_((int, char *[]));
 static int NEAR getyesno __P_((int, char *[]));
 #endif	/* FD >= 2 */
+#if	!MSDOS && (FD >= 2)
+static int NEAR savetty __P_((int, char *[]));
+#endif
 #ifdef	_NOORIGSHELL
 static int NEAR printenv __P_((int, char *[]));
 static int NEAR setalias __P_((int, char *[]));
@@ -140,7 +146,7 @@ static int NEAR loadsource __P_((int, char *[]));
 #endif
 
 static CONST char *builtinerrstr[] = {
-	"",
+	NULL,
 #define	ER_FEWMANYARG	1
 	"Too few or many arguments",
 #define	ER_OUTOFLIMIT	2
@@ -204,6 +210,9 @@ static CONST builtintable builtinlist[] = {
 	{getinputstr,	BL_READLINE},
 	{getyesno,	BL_YESNO},
 #endif	/* FD >= 2 */
+#if	!MSDOS && (FD >= 2)
+	{savetty,	BL_SAVETTY},
+#endif
 #ifdef	_NOORIGSHELL
 # if	FD >= 2
 	{printenv,	BL_SET},
@@ -240,7 +249,8 @@ int n;
 	if (argv && argv[0]) fprintf2(stderr, "%k: ", argv[0]);
 	if (s) fprintf2(stderr, "%k: ", s);
 	fprintf2(stderr, "%s.",
-		(n >= 0) ? builtinerrstr[n] : strerror2(duperrno));
+		(n >= 0 && builtinerrstr[n])
+			? builtinerrstr[n] : strerror2(duperrno));
 	fputnl(stderr);
 }
 
@@ -1324,7 +1334,7 @@ int n;
 	freemacro(bindlist[n].d_func);
 	if (bindlist[n].key >= K_F(1) && bindlist[n].key <= K_F(10)) {
 		free(helpindex[bindlist[n].key - K_F(1)]);
-		helpindex[bindlist[n].key - K_F(1)] = strdup2("");
+		helpindex[bindlist[n].key - K_F(1)] = strdup2(nullstr);
 	}
 	memmove((char *)&(bindlist[n]), (char *)&(bindlist[n + 1]),
 		(MAXBINDTABLE - n) * sizeof(bindtable));
@@ -1592,6 +1602,7 @@ devinfo *devp;
 		}
 		free(sp);
 		free(drvlist);
+		free(devp -> name);
 		return(no);
 	}
 	fdtype[no].offset = 0;
@@ -2160,7 +2171,8 @@ char *argv[];
 
 		i = n;
 		for (; n < argc; n++) if (!strchr(argv[n], '=')) break;
-		if ((f = parsehist((n < argc) ? argv[n] : "!", NULL)) < 0) {
+		f = parsehist((n < argc) ? argv[n] : "!", NULL, '\0');
+		if (f < 0) {
 			builtinerror(argv, argv[n], ER_EVENTNOFOUND);
 			entryhist(0, tmp, 0);
 			free(tmp);
@@ -2203,12 +2215,12 @@ char *argv[];
 	if (!editor) editor = EDITOR;
 
 	if (list) {
-		f = parsehist((n < argc) ? argv[n] : "-16", NULL);
-		l = parsehist((n + 1 < argc) ? argv[n + 1] : "!", NULL);
+		f = parsehist((n < argc) ? argv[n] : "-16", NULL, '\0');
+		l = parsehist((n + 1 < argc) ? argv[n + 1] : "!", NULL, '\0');
 	}
 	else {
-		f = parsehist((n < argc) ? argv[n] : "!", NULL);
-		l = (n + 1 < argc) ? parsehist(argv[n + 1], NULL) : f;
+		f = parsehist((n < argc) ? argv[n] : "!", NULL, '\0');
+		l = (n + 1 < argc) ? parsehist(argv[n + 1], NULL, '\0') : f;
 	}
 	if (f < 0 || l < 0) {
 		if (f < 0) builtinerror(argv, argv[n], ER_EVENTNOFOUND);
@@ -2242,7 +2254,7 @@ char *argv[];
 		}
 		if (!(fp = Xfdopen(fd, "w"))) {
 			builtinerror(argv, path, -1);
-			Xclose(fd);
+			VOID_C Xclose(fd);
 			rmtmpfile(path);
 			free(tmp);
 			return(-1);
@@ -2681,7 +2693,7 @@ char *argv[];
 	}
 	if (!(wastty = isttyiomode)) ttyiomode(1);
 	duppromptstr = promptstr;
-	promptstr = (argc >= 2) ? argv[1] : "";
+	promptstr = (argc >= 2) ? argv[1] : nullstr;
 	s = inputstr(NULL, 0, -1, NULL, -1);
 	promptstr = duppromptstr;
 	if (!wastty) stdiomode();
@@ -2705,7 +2717,7 @@ char *argv[];
 		builtinerror(argv, NULL, ER_NOTDUMBTERM);
 		return(-1);
 	}
-	if (argc < 2) s = "";
+	if (argc < 2) s = nullstr;
 	else s = argv[1];
 
 	if (!(wastty = isttyiomode)) ttyiomode(1);
@@ -2717,6 +2729,38 @@ char *argv[];
 }
 #endif	/* FD >= 2 */
 
+#if	!MSDOS
+int savestdio(reset)
+int reset;
+{
+	int n;
+
+	if (isttyiomode) return(0);
+
+	n = savettyio(reset);
+# ifndef	_NOPTY
+	sendparent(TE_SAVETTYIO, reset);
+# endif
+	return(n);
+}
+
+# if	FD >= 2
+static int NEAR savetty(argc, argv)
+int argc;
+char *argv[];
+{
+	int n;
+
+	n = 0;
+	if (argc > 1 && argv[1][0] == '-'
+	&& argv[1][1] == 'n' && !(argv[1][2]))
+		n++;
+
+	return(savestdio(n));
+}
+# endif	/* FD >= 2 */
+#endif	/* !MSDOS */
+
 #ifdef	_NOORIGSHELL
 static int NEAR printenv(argc, argv)
 int argc;
@@ -3230,7 +3274,7 @@ char *argv[];
 		builtinerror(argv, NULL, ER_NOTDUMBTERM);
 		return(-1);
 	}
-	if (fd_restricted && (funclist[n].status & RESTRICT)) {
+	if (fd_restricted && (funclist[n].status & FN_RESTRICT)) {
 		fprintf2(stderr, "%s: %k", argv[0], RESTR_K);
 		fputnl(stderr);
 		return(RET_NOTICE);
@@ -3240,6 +3284,12 @@ char *argv[];
 		fputnl(stderr);
 		return(RET_NOTICE);
 	}
+#ifndef	_NOPTY
+	if (parentfd >= 0) {
+		sendparent(TE_INTERNAL, n, argv[1]);
+		return(RET_SUCCESS);
+	}
+#endif
 	ttyiomode(0);
 	internal_status = (*funclist[n].func)(argv[1]);
 	locate(0, n_line - 1);
@@ -3258,7 +3308,7 @@ int flags;
 
 	n = -1;
 	command = skipspace(command);
-	internal_status = -2;
+	internal_status = FNC_FAIL;
 	if (!(argc = getargs(command, &argv))) {
 		freevar(argv);
 		return(-1);
