@@ -93,6 +93,20 @@ extern int ptymode;
 extern char *ptyterm;
 extern int ptymenukey;
 #endif
+#ifndef	_NOLOGGING
+extern char *logfile;
+extern int logsize;
+# ifndef	NOSYSLOG
+extern int usesyslog;
+# endif
+extern int loglevel;
+# ifndef	NOUID
+extern int rootloglevel;
+# endif
+#endif	/* !_NOLOGGING */
+#if	FD >= 2
+extern int thruargs;
+#endif
 extern int wheader;
 extern char fullpath[];
 extern char *origpath;
@@ -165,6 +179,8 @@ typedef struct _envtable {
 	u_char type;
 } envtable;
 
+#define	env_str(n)	(&(envlist[n].env[FDESIZ]))
+#define	fdenv_str(n)	(envlist[n].env)
 #ifdef	FORCEDSTDC
 #define	def_str(n)	(envlist[n].def.str)
 #define	def_num(n)	(envlist[n].def.num)
@@ -196,8 +212,9 @@ typedef struct _envtable {
 #define	T_KIN		14
 #define	T_KOUT		15
 #define	T_KNAM		16
-#define	T_OCTAL		17
-#define	T_KEYCODE	18
+#define	T_KTERM		17
+#define	T_OCTAL		18
+#define	T_KEYCODE	19
 
 #ifndef	_NOKANJIFCONV
 typedef struct _pathtable {
@@ -393,6 +410,21 @@ static CONST envtable envlist[] = {
 	{"FD_PTYTERM", &ptyterm, DEFVAL(PTYTERM), PTYTM_E, T_CHARP},
 	{"FD_PTYMENUKEY", &ptymenukey, DEFVAL(PTYMENUKEY), PTYKY_E, T_KEYCODE},
 #endif
+#ifndef	_NOLOGGING
+	{"FD_LOGFILE", &logfile, DEFVAL(LOGFILE), LGFIL_E, T_PATH},
+	{"FD_LOGSIZE", &logsize, DEFVAL(LOGSIZE), LGSIZ_E, T_INT},
+# ifndef	NOSYSLOG
+	{"FD_USESYSLOG", &usesyslog, DEFVAL(USESYSLOG), USYLG_E, T_BOOL},
+# endif
+	{"FD_LOGLEVEL", &loglevel, DEFVAL(LOGLEVEL), LGLVL_E, T_INT},
+# ifndef	NOUID
+	{"FD_ROOTLOGLEVEL", &rootloglevel,
+		DEFVAL(ROOTLOGLEVEL), RLGLV_E, T_INT},
+# endif
+#endif	/* !_NOLOGGING */
+#if	FD >= 2
+	{"FD_THRUARGS", &thruargs, DEFVAL(THRUARGS), THARG_E, T_BOOL},
+#endif
 #if	!defined (_NOKANJICONV) || (defined (FD) && !defined (_NODOSDRIVE))
 	{"FD_UNICODEBUFFER", &unicodebuffer,
 		DEFVAL(UNICODEBUFFER), UNBF_E, T_BOOL},
@@ -407,6 +439,10 @@ static CONST envtable envlist[] = {
 #ifndef	_NOKANJICONV
 	{"FD_INPUTKCODE", &inputkcode, DEFVAL(NOCNV), IPKC_E, T_KIN},
 #endif
+#if	!defined (_NOKANJICONV) && !defined (_NOPTY)
+	{"FD_PTYINKCODE", &ptyinkcode, DEFVAL(NOCNV), PIKC_E, T_KTERM},
+	{"FD_PTYOUTKCODE", &ptyoutkcode, DEFVAL(NOCNV), POKC_E, T_KTERM},
+#endif
 #ifndef	_NOKANJIFCONV
 	{"FD_FNAMEKCODE", &fnamekcode, DEFVAL(NOCNV), FNKC_E, T_KNAM},
 	{"FD_SJISPATH", &sjispath, DEFVAL(SJISPATH), SJSP_E, T_PATHS},
@@ -501,7 +537,7 @@ int no;
 	char *cp;
 	int n;
 
-	cp = getenv2(envlist[no].env);
+	cp = getenv2(fdenv_str(no));
 	switch (envlist[no].type) {
 		case T_BOOL:
 			if (!cp) n = def_num(no);
@@ -599,8 +635,9 @@ int no;
 		case T_KIN:
 		case T_KOUT:
 		case T_KNAM:
-			*((int *)(envlist[no].var)) =
-				getlang(cp, envlist[no].type - T_KIN);
+		case T_KTERM:
+			n = (1 << (envlist[no].type - T_KIN));
+			*((int *)(envlist[no].var)) = getlang(cp, n);
 			break;
 #endif	/* !_NOKANJICONV || (!_NOENGMES && !NOJPNMES) */
 #if	FD >= 2
@@ -635,7 +672,10 @@ int stable;
 
 	if (!stable) /*EMPTY*/;
 	else if (pp -> lang == NOCNV) {
-		pp -> last = NULL;
+		if (!(pp -> flags & P_STABLE)) {
+			if (pp -> last) free(pp -> last);
+			pp -> last = NULL;
+		}
 		stable = 0;
 	}
 	else if (pp -> flags & P_STABLE) /*EMPTY*/;
@@ -668,7 +708,10 @@ pathtable *pp;
 		}
 		if (kanjierrno) pp -> lang = DEFCODE;
 	}
-	pp -> last = strdup2(path);
+	if (!(pp -> flags & P_STABLE)) {
+		if (pp -> last) free(pp -> last);
+		pp -> last = strdup2(path);
+	}
 }
 
 static VOID NEAR savepathlang(VOID_A)
@@ -705,6 +748,8 @@ static VOID NEAR evalpathlang(VOID_A)
 # endif	/* !_NOSPLITWIN */
 # ifndef	_NOPTY
 	changekcode();
+	changeinkcode();
+	changeoutkcode();
 # endif
 }
 #endif	/* !_NOKANJIFCONV */
@@ -748,6 +793,9 @@ VOID evalenv(VOID_A)
 #endif
 	evalheader();
 	errno = duperrno;
+#if	!MSDOS && !defined (_NOORIGSHELL)
+	if (autosavetty) savestdio(0);
+#endif
 }
 
 #ifdef	DEBUG
@@ -755,6 +803,12 @@ VOID freeenvpath(VOID_A)
 {
 	int i;
 
+# ifndef	_NOKANJIFCONV
+	for (i = 0; i < PATHLISTSIZ; i++) {
+		if (pathlist[i].last) free(pathlist[i].last);
+		pathlist[i].last = NULL;
+	}
+# endif
 	for (i = 0; i < ENVLISTSIZ; i++) switch (envlist[i].type) {
 		case T_PATH:
 		case T_PATHS:
@@ -766,7 +820,7 @@ VOID freeenvpath(VOID_A)
 			break;
 	}
 }
-#endif
+#endif	/* DEBUG */
 
 #ifndef	_NOCUSTOMIZE
 static int NEAR custputs(s)
@@ -800,7 +854,7 @@ int y, w;
 {
 	Xlocate(0, y);
 	Xputterm(T_STANDOUT);
-	Xcprintf2("%*s", w, "");
+	cputspace(w);
 	Xputterm(END_STANDOUT);
 }
 
@@ -1057,9 +1111,8 @@ char **list;
 		savepathlang();
 #endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			setenv2(envlist[i].env, list[i * 2], 0);
-			setenv2(&(envlist[i].env[FDESIZ]),
-				list[i * 2 + 1], 0);
+			setenv2(fdenv_str(i), list[i * 2], 0);
+			setenv2(env_str(i), list[i * 2 + 1], 0);
 			_evalenv(i);
 		}
 #ifndef	_NOKANJIFCONV
@@ -1070,10 +1123,8 @@ char **list;
 	else {
 		list = (char **)malloc2(ENVLISTSIZ * 2 * sizeof(char *));
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			list[i * 2] = strdup2(getshellvar(envlist[i].env, -1));
-			list[i * 2 + 1] =
-				strdup2(getshellvar(&(envlist[i].env[FDESIZ]),
-					-1));
+			list[i * 2] = strdup2(getshellvar(fdenv_str(i), -1));
+			list[i * 2 + 1] = strdup2(getshellvar(env_str(i), -1));
 		}
 	}
 
@@ -1089,9 +1140,9 @@ static VOID NEAR cleanupenv(VOID_A)
 	savepathlang();
 #endif
 	for (i = 0; i < ENVLISTSIZ; i++) {
-		setenv2(envlist[i].env, NULL, 0);
+		setenv2(fdenv_str(i), NULL, 0);
 		cp = (envlist[i].type == T_CHARP) ? def_str(i) : NULL;
-		setenv2(&(envlist[i].env[FDESIZ]), cp, 0);
+		setenv2(env_str(i), cp, 0);
 		_evalenv(i);
 	}
 #ifndef	_NOKANJIFCONV
@@ -1249,6 +1300,7 @@ int no;
 		case T_KIN:
 		case T_KOUT:
 		case T_KNAM:
+		case T_KTERM:
 			str[NOCNV] = VNCNV_K;
 			str[ENG] = VENG_K;
 #  ifndef	_NOKANJICONV
@@ -1275,18 +1327,18 @@ int no;
 # endif
 # ifndef	_NOPTY
 		case T_KEYCODE:
-			cp = getenv2(envlist[no].env);
+			cp = getenv2(fdenv_str(no));
 			break;
 # endif
 		default:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
 			break;
 	}
 
-	if (!getenv2(envlist[no].env))
+	if (!getenv2(fdenv_str(no)))
 		cp = new = strcatalloc(strdup2(cp),
 			(cp && *cp) ? VDEF_K : VUDEF_K);
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, cp);
+	cputstr(MAXCUSTVAL, cp);
 	n = strlen2(cp);
 	if (n > MAXCUSTVAL - 1) n = MAXCUSTVAL - 1;
 	if (new) free(new);
@@ -1303,7 +1355,7 @@ int no;
 	for (n = 0; n < MAXSELECTSTRS; n++) val[n] = n;
 
 	new = NULL;
-	env = &(envlist[no].env[FDESIZ]);
+	env = env_str(no);
 	switch (envlist[no].type) {
 		case T_BOOL:
 			n = (*((int *)(envlist[no].var))) ? 1 : 0;
@@ -1344,8 +1396,8 @@ int no;
 			break;
 		case T_PATH:
 		case T_PATHS:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
-			new = inputcustenvstr(env, 1, cp, 1);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
+			new = inputcustenvstr(env, 1, cp, HST_PATH);
 			if (new == (char *)-1) return(0);
 			cp = new;
 			break;
@@ -1508,6 +1560,7 @@ int no;
 		case T_KIN:
 		case T_KNAM:
 		case T_KOUT:
+		case T_KTERM:
 			tmp = 0;
 			str[tmp] = VNCNV_K;
 			val[tmp++] = NOCNV;
@@ -1526,29 +1579,17 @@ int no;
 			val[tmp++] = JIS8;
 			str[tmp] = "ISO-2022-JP";
 			val[tmp++] = JUNET;
+			str[tmp] = "Hex";
+			val[tmp++] = HEX;
 			str[tmp] = "UTF-8";
 			val[tmp++] = UTF8;
 #  endif	/* !_NOKANJICONV */
 			str[tmp] = VUSET_K;
 			val[tmp++] = -1;
-			p = envlist[no].type - T_KIN;
+			p = (1 << (envlist[no].type - T_KIN));
 			for (n = 0; n < tmp; n++) {
-				if (val[n] < 0) continue;
-				switch (kanjiiomode[val[n]]) {
-					case L_INPUT:
-						if (p != L_INPUT) continue;
-						break;
-					case L_OUTPUT:
-						if (p == L_OUTPUT) continue;
-						break;
-					case L_FNAME:
-						if (p == L_FNAME) continue;
-						break;
-					default:
-						continue;
-/*NOTREACHED*/
-						break;
-				}
+				if (val[n] < 0 || (kanjiiomode[val[n]] & p))
+					continue;
 				tmp--;
 				memmove(&(str[n]), &(str[n + 1]),
 					(tmp - n) * sizeof(char *));
@@ -1559,12 +1600,18 @@ int no;
 			if (!tmp) return(0);
 			n = *((int *)(envlist[no].var));
 #  ifndef	_NOKANJICONV
-			if (n != O_JIS7 && n != O_JIS8 && n != O_JUNET
-			&& n != M_UTF8)
-				p = 0;
-			else {
-				p = 1;
-				n--;
+			switch (n) {
+				case O_JIS7:
+				case O_JIS8:
+				case O_JUNET:
+				case CAP:
+				case M_UTF8:
+					p = 1;
+					n--;
+					break;
+				default:
+					p = 0;
+					break;
 			}
 #  endif	/* !_NOKANJICONV */
 			envcaption(env);
@@ -1574,24 +1621,33 @@ int no;
 				break;
 			}
 #  ifndef	_NOKANJICONV
-			if (n >= JIS7 && n <= JUNET) {
-				str[0] = VNJIS_K;
-				str[1] = VOJIS_K;
-				val[0] = 0;
-				val[1] = 1;
-				if (noselect(&p, 2, 64, str, val)) return(0);
-				n += p;
+			switch (n) {
+				case JIS7:
+				case JIS8:
+				case JUNET:
+					str[0] = VNJIS_K;
+					str[1] = VOJIS_K;
+					break;
+				case HEX:
+					str[0] = "HEX";
+					str[1] = "CAP";
+					break;
+				case UTF8:
+					str[0] = VUTF8_K;
+					str[1] = VUTFM_K;
+					break;
+				default:
+					tmp = -1;
+					break;
 			}
-			else if (n == UTF8) {
-				str[0] = VUTF8_K;
-				str[1] = VUTFM_K;
+			if (tmp >= 0) {
 				val[0] = 0;
 				val[1] = 1;
 				if (noselect(&p, 2, 64, str, val)) return(0);
 				n += p;
 			}
 #  endif	/* !_NOKANJICONV */
-			str[NOCNV] = "";
+			str[NOCNV] = nullstr;
 #  if	!defined (_NOENGMES) && !defined (_NOJPNMES)
 			str[ENG] = "C";
 #  endif
@@ -1629,7 +1685,7 @@ int no;
 # endif	/* FD >= 2 */
 # ifndef	_NOPTY
 		case T_KEYCODE:
-			if (!(cp = getenv2(envlist[no].env))) cp = "";
+			if (!(cp = getenv2(fdenv_str(no)))) cp = nullstr;
 			new = inputcustenvstr(env, 0, cp, -1);
 			if (new == (char *)-1) return(0);
 			if (new && getkeycode(new, 0) < 0) {
@@ -1641,16 +1697,17 @@ int no;
 			break;
 # endif	/* !_NOPTY */
 		default:
-			if (!(cp = getenv2(envlist[no].env))) cp = def_str(no);
+			if (!(cp = getenv2(fdenv_str(no)))) cp = def_str(no);
 			new = inputcustenvstr(env, 0, cp, -1);
 			if (new == (char *)-1) return(0);
 			cp = new;
 			break;
 	}
 
-	if (getshellvar(envlist[no].env, -1)) setenv2(envlist[no].env, cp, 0);
-	else setenv2(env, cp, 0);
+	if (getshellvar(fdenv_str(no), -1)) n = setenv2(fdenv_str(no), cp, 0);
+	else n = setenv2(env, cp, 0);
 	if (new) free(new);
+	if (n < 0) warning(-1, env);
 #ifndef	_NOKANJIFCONV
 	savepathlang();
 #endif
@@ -1672,11 +1729,11 @@ FILE *fp;
 
 	for (i = n = 0; i < ENVLISTSIZ; i++) {
 		if ((!flaglist || !(flaglist[i] & 2))
-		&& (cp = getshellvar(&(envlist[i].env[FDESIZ]), -1))
+		&& (cp = getshellvar(env_str(i), -1))
 		&& (envlist[i].type != T_CHARP || strenvcmp(cp, def_str(i))))
 			n++;
 		if ((!flaglist || !(flaglist[i] & 1))
-		&& getshellvar(envlist[i].env, -1))
+		&& getshellvar(fdenv_str(i), -1))
 			n++;
 	}
 	if (!n || !fp) return(n);
@@ -1685,16 +1742,16 @@ FILE *fp;
 	fputs("# shell variables definition\n", fp);
 	for (i = 0; i < ENVLISTSIZ; i++) {
 		if ((!flaglist || !(flaglist[i] & 2))
-		&& (cp = getshellvar(&(envlist[i].env[FDESIZ]), -1))
+		&& (cp = getshellvar(env_str(i), -1))
 		&& (envlist[i].type != T_CHARP || strenvcmp(cp, def_str(i)))) {
 			cp = killmeta(cp);
-			fprintf2(fp, "%s=%s\n", &(envlist[i].env[FDESIZ]), cp);
+			fprintf2(fp, "%s=%s\n", env_str(i), cp);
 			free(cp);
 		}
 		if ((!flaglist || !(flaglist[i] & 1))
-		&& (cp = getshellvar(envlist[i].env, -1))) {
+		&& (cp = getshellvar(fdenv_str(i), -1))) {
 			cp = killmeta(cp);
-			fprintf2(fp, "%s=%s\n", envlist[i].env, cp);
+			fprintf2(fp, "%s=%s\n", fdenv_str(i), cp);
 			free(cp);
 		}
 	}
@@ -1713,10 +1770,8 @@ FILE *fp;
 
 	for (n = 0; argv[n]; n++) {
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
-			if (!strnenvcmp(argv[n], ident, len[n])) break;
-			ident += FDESIZ;
-			if (!strnenvcmp(argv[n], ident, len[n])) break;
+			if (!strnenvcmp(argv[n], fdenv_str(i), len[n])) break;
+			if (!strnenvcmp(argv[n], env_str(i), len[n])) break;
 		}
 		if (i < ENVLISTSIZ) break;
 	}
@@ -1729,12 +1784,12 @@ FILE *fp;
 		f = 0;	/* fake for -Wuninitialized */
 #  endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
+			ident = fdenv_str(i);
 			if (!strnenvcmp(argv[n], ident, len[n])) {
 				f = 1;
 				break;
 			}
-			ident += FDESIZ;
+			ident = env_str(i);
 			if (!strnenvcmp(argv[n], ident, len[n])) {
 				f = 2;
 				break;
@@ -1794,10 +1849,8 @@ FILE *fp;
 
 	for (n = 1; argv[n]; n++) {
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
-			if (!strenvcmp(argv[n], ident)) break;
-			ident += FDESIZ;
-			if (!strenvcmp(argv[n], ident)) break;
+			if (!strenvcmp(argv[n], fdenv_str(i))) break;
+			if (!strenvcmp(argv[n], env_str(i))) break;
 		}
 		if (i < ENVLISTSIZ) break;
 	}
@@ -1810,12 +1863,12 @@ FILE *fp;
 		f = 0;	/* fake for -Wuninitialized */
 #  endif
 		for (i = 0; i < ENVLISTSIZ; i++) {
-			ident = envlist[i].env;
+			ident = fdenv_str(i);
 			if (!strenvcmp(argv[n], ident)) {
 				f = 1;
 				break;
 			}
-			ident += FDESIZ;
+			ident = env_str(i);
 			if (!strenvcmp(argv[n], ident)) {
 				f = 2;
 				break;
@@ -1897,7 +1950,7 @@ int no;
 	int len, width;
 
 	if (bindlist[no].key < 0) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	if (bindlist[no].f_func < FUNCLISTSIZ)
@@ -1910,14 +1963,13 @@ int no;
 	else cp2 = NULL;
 	if (!cp2) {
 		width = MAXCUSTVAL;
-		Xcprintf2("%-*.*k", width, width, cp1);
+		cputstr(width, cp1);
 	}
 	else {
 		width = (MAXCUSTVAL - 1) / 2;
-		Xcprintf2("%-*.*k", width, width, cp1);
+		cputstr(width, cp1);
 		putsep();
-		Xcprintf2("%-*.*k",
-			MAXCUSTVAL - 1 - width, MAXCUSTVAL - 1 - width, cp2);
+		cputstr(MAXCUSTVAL - 1 - width, cp2);
 	}
 	len = strlen2(cp1);
 	if (len > --width) len = width;
@@ -2031,7 +2083,7 @@ int no;
 				cp = NULL;
 			else cp = macrolist[bindlist[no].f_func - FUNCLISTSIZ];
 			buf = asprintf3(BNDFC_K, str);
-			cp = inputcuststr(buf, 0, cp, 0);
+			cp = inputcuststr(buf, 0, cp, HST_COM);
 			free(buf);
 			if (!cp);
 			else if (!*cp) {
@@ -2068,7 +2120,7 @@ int no;
 			if (bindlist[no].key < 0 || i < FUNCLISTSIZ) cp = NULL;
 			else cp = macrolist[i - FUNCLISTSIZ];
 			buf = asprintf3(BNDDC_K, str);
-			cp = inputcuststr(buf, 0, cp, 0);
+			cp = inputcuststr(buf, 0, cp, HST_COM);
 			free(buf);
 			if (!cp);
 			else if (!*cp) {
@@ -2217,11 +2269,11 @@ int no;
 
 	key.code = keyseqlist[no];
 	if (getkeyseq(&key) < 0 || !(key.len)) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	cp = encodestr(key.str, key.len);
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, cp);
+	cputstr(MAXCUSTVAL, cp);
 	len = strlen2(cp);
 	if (len > MAXCUSTVAL - 1) len = MAXCUSTVAL - 1;
 	free(cp);
@@ -2472,19 +2524,19 @@ int no;
 	int len, width;
 
 	if (no >= maxlaunch) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	if (!launchlist[no].format) {
 		width = MAXCUSTVAL;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].comm);
+		cputstr(width, launchlist[no].comm);
 	}
 	else {
 		width = (MAXCUSTVAL - 1 - 6) / 2;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].comm);
+		cputstr(width, launchlist[no].comm);
 		putsep();
 		width = MAXCUSTVAL - 1 - 6 - width;
-		Xcprintf2("%-*.*k", width, width, launchlist[no].format[0]);
+		cputstr(width, launchlist[no].format[0]);
 		putsep();
 		Xcprintf2("%-2d", (int)(launchlist[no].topskip));
 		putsep();
@@ -2516,7 +2568,7 @@ launchtable *list;
 	}
 	Xcprintf2("%.*k", len, list -> ext + 1);
 	putsep();
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, list -> comm);
+	cputstr(MAXCUSTVAL, list -> comm);
 	putsep();
 	fillline(yy + y++, n_column);
 	nf = ni = ne = 0;
@@ -2546,8 +2598,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> format[nf]);
+			cputstr(MAXCUSTVAL, list -> format[nf]);
 			putsep();
 			nf++;
 		}
@@ -2565,8 +2616,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> lignore[ni]);
+			cputstr(MAXCUSTVAL, list -> lignore[ni]);
 			putsep();
 			ni++;
 		}
@@ -2584,8 +2634,7 @@ launchtable *list;
 				Xputterm(END_STANDOUT);
 				putsep();
 			}
-			Xcprintf2("%-*.*k",
-				MAXCUSTVAL, MAXCUSTVAL, list -> lerror[ne]);
+			cputstr(MAXCUSTVAL, list -> lerror[ne]);
 			putsep();
 			ne++;
 		}
@@ -2799,7 +2848,8 @@ int no;
 		list.topskip = launchlist[no].topskip;
 		list.bottomskip = launchlist[no].bottomskip;
 
-		list.comm = inputcuststr(LNCHC_K, 0, launchlist[no].comm, 0);
+		list.comm = inputcuststr(LNCHC_K, 0,
+			launchlist[no].comm, HST_COM);
 		if (!(list.comm)) {
 			freelaunch(&list);
 			return(-1);
@@ -3021,18 +3071,13 @@ int no;
 	int len, width;
 
 	if (no >= maxarchive) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	width = (MAXCUSTVAL - 1) / 2;
-	if (!archivelist[no].p_comm) Xcprintf2("%*s", width, "");
-	else Xcprintf2("%-*.*k", width, width, archivelist[no].p_comm);
+	cputstr(width, archivelist[no].p_comm);
 	putsep();
-	if (!archivelist[no].u_comm)
-		Xcprintf2("%*s", MAXCUSTVAL - 1 - width, "");
-	else Xcprintf2("%-*.*k",
-		MAXCUSTVAL - 1 - width, MAXCUSTVAL - 1 - width,
-		archivelist[no].u_comm);
+	cputstr(MAXCUSTVAL - 1 - width, archivelist[no].u_comm);
 	len = (archivelist[no].p_comm) ? strlen2(archivelist[no].p_comm) : 0;
 	if (len > --width) len = width;
 
@@ -3079,7 +3124,7 @@ int no;
 
 	for (;;) {
 		list.p_comm = inputcuststr(PACKC_K, 0,
-			archivelist[no].p_comm, 0);
+			archivelist[no].p_comm, HST_COM);
 		if (!(list.p_comm)) {
 			free(list.ext);
 			return(0);
@@ -3090,7 +3135,7 @@ int no;
 		}
 
 		list.u_comm = inputcuststr(UPCKC_K, 0,
-			archivelist[no].u_comm, 0);
+			archivelist[no].u_comm, HST_COM);
 		if (!(list.u_comm)) {
 			if (list.p_comm) free(list.p_comm);
 			continue;
@@ -3277,11 +3322,11 @@ int no;
 	int i, len, w1, w2, width;
 
 	if (!fdtype[no].name) {
-		Xcprintf2("%*s", MAXCUSTVAL, "");
+		cputspace(MAXCUSTVAL);
 		return(0);
 	}
 	width = (MAXCUSTVAL - 1) / 2;
-	Xcprintf2("%-*.*k", width, width, fdtype[no].name);
+	cputstr(width, fdtype[no].name);
 	putsep();
 
 	w1 = MAXCUSTVAL - 1 - width;
@@ -3294,7 +3339,7 @@ int no;
 		i = strlen(buf);
 		snprintf2(&(buf[i]), sizeof(buf) - i, " #offset=%'Ld",
 			fdtype[no].offset / fdtype[no].sect);
-		Xcprintf2("%-*.*k", w1, w1, buf);
+		cputstr(w1, buf);
 	}
 	else
 #  endif	/* HDDMOUNT */
@@ -3321,8 +3366,7 @@ int no;
 				break;
 		}
 		w1 -= (w2 + 1) * 3;
-		if (i >= MEDIADESCRSIZ) Xcprintf2("%*s", w1, "");
-		else Xcprintf2("%-*.*k", w1, w1, mediadescr[i].name);
+		cputstr(w1, (i < MEDIADESCRSIZ) ? mediadescr[i].name : NULL);
 	}
 	len = strlen2(fdtype[no].name);
 	if (len > --width) len = width;
@@ -3375,7 +3419,8 @@ int no;
 	for (;;) {
 		if (!(dev.name)) {
 			cp = asprintf3(DRDEV_K, dev.drive);
-			dev.name = inputcuststr(cp, 1, fdtype[no].name, 1);
+			dev.name = inputcuststr(cp, 1,
+				fdtype[no].name, HST_PATH);
 			free(cp);
 			if (!(dev.name)) return(0);
 			else if (!*(dev.name)) {
@@ -3616,7 +3661,7 @@ int no;
 # else
 	mes[4] = COVWR_K;
 # endif
-	Xcprintf2("%-*.*k", MAXCUSTVAL, MAXCUSTVAL, mes[no]);
+	cputstr(MAXCUSTVAL, mes[no]);
 	len = strlen2(mes[no]);
 	if (len > MAXCUSTVAL - 1) len = MAXCUSTVAL - 1;
 
@@ -3662,7 +3707,7 @@ char *file;
 	}
 	if (!(fpout = Xfdopen(fd, "w"))) {
 		warning(-1, path);
-		Xclose(fd);
+		VOID_C Xclose(fd);
 		Xunlink(path);
 		if (fpin) Xfclose(fpin);
 		return(-1);
@@ -3918,8 +3963,8 @@ int no;
 # endif
 			break;
 		case 2:
-			if (!(file = inputcuststr(FLOAD_K, 1, FD_RCFILE, 1)))
-				break;
+			file = inputcuststr(FLOAD_K, 1, FD_RCFILE, HST_PATH);
+			if (!file) break;
 			if (!*file) {
 				free(file);
 				break;
@@ -3938,8 +3983,8 @@ int no;
 			}
 			break;
 		case 3:
-			if (!(file = inputcuststr(FSAVE_K, 1, FD_RCFILE, 1)))
-				break;
+			file = inputcuststr(FSAVE_K, 1, FD_RCFILE, HST_PATH);
+			if (!file) break;
 			done = 1;
 			file = evalpath(file, 0);
 # ifdef	FAKEUNINIT
@@ -3977,8 +4022,8 @@ int no;
 			break;
 		case 4:
 # ifndef	_NOORIGSHELL
-			if (!(file = inputcuststr(FOVWR_K, 1, FD_RCFILE, 1)))
-				break;
+			file = inputcuststr(FOVWR_K, 1, FD_RCFILE, HST_PATH);
+			if (!file) break;
 			done = 1;
 			file = evalpath(file, 0);
 			if (!*file
@@ -4014,7 +4059,7 @@ int no, y, isstandout;
 	cp = NEWET_K;
 	switch (custno) {
 		case 0:
-			cp = &(envlist[no].env[FDESIZ]);
+			cp = env_str(no);
 			break;
 		case 1:
 			if (bindlist[no].key >= 0)
@@ -4070,7 +4115,7 @@ int no, y, isstandout;
 			cp = name[no];
 			break;
 		default:
-			cp = "";
+			cp = nullstr;
 			break;
 	}
 
@@ -4133,8 +4178,7 @@ static VOID NEAR dispcust(VOID_A)
 					Xcprintf2("%.*k", MAXCUSTVAL, NIMPL_K);
 				if (cs_len[i - start] > MAXCUSTVAL - 1)
 					cs_len[i - start] = MAXCUSTVAL - 1;
-				else Xcprintf2("%*s",
-					MAXCUSTVAL - cs_len[i - start], "");
+				else cputspace(MAXCUSTVAL - cs_len[i - start]);
 				break;
 		}
 		putsep();
