@@ -1,13 +1,19 @@
 /*
  *	dosdisk.h
  *
- *	Type Definition for "dosdisk.c"
+ *	type definitions for "dosdisk.c"
  */
 
 #if	!MSDOS
 #include <sys/param.h>
 #endif
 
+#ifndef	__SYS_TYPES_STAT_H_
+#define	__SYS_TYPES_STAT_H_
+#include <sys/types.h>
+#include <sys/stat.h>
+#endif
+
 #define	DOSDIRENT	32
 #define	SECTCACHESIZE	32
 #define	SECTSIZE	{512, 1024, 256, 128, 2048, 4096}
@@ -26,12 +32,12 @@
 #define	INHIBITNAME	{"CON     ", "AUX     ", "PRN     ", \
 			 "CLOCK$  ", "CONFIG$ ", "NUL     "}
 #define	INHIBITCOM	"COM"
-#define	INHIBITCOMMAX	4
+#define	INHIBITCOMMAX	9	/* NT architecture will avoid COM5-COM9 */
 #define	INHIBITLPT	"LPT"
-#define	INHIBITLPTMAX	9
+#define	INHIBITLPTMAX	9	/* Windows will avoid LPT4-LPT9 */
 
 #define	MAXDRIVEENTRY	32
-#define	DOSFDOFFSET	(1 << (8 * sizeof(int) - 2))
+#define	DOSFDOFFSET	(1 << (8 * (int)sizeof(int) - 2))
 #ifdef	NOFILE
 #define	DOSNOFILE	NOFILE
 #else
@@ -46,6 +52,10 @@
 #define	d_fileno	d_ino
 #endif
 
+#ifdef	NODRECLEN
+#define	d_reclen	d_fd
+#endif
+
 #ifndef	L_SET
 # ifdef	SEEK_SET
 # define	L_SET	SEEK_SET
@@ -81,9 +91,9 @@
 #endif
 
 #ifdef	USELLSEEK
-typedef long long	off64_t;
+typedef long long	l_off_t;
 #else
-#define	off64_t		off_t
+typedef off_t		l_off_t;
 #endif
 
 typedef struct _bpb_t {
@@ -143,7 +153,7 @@ typedef struct _cache_t {
 	char path[MAXPATHLEN];
 	dent_t dent;
 	long clust;
-	u_short offset;
+	long offset;
 } cache_t;
 
 #if	!MSDOS
@@ -154,7 +164,7 @@ typedef struct _devinfo {
 	u_short sect;
 	u_short cyl;
 # ifdef	HDDMOUNT
-	off64_t offset;
+	l_off_t offset;
 # endif
 } devinfo;
 #endif
@@ -166,7 +176,7 @@ typedef struct _devstat {
 	u_short sectsize;	/* ch_sect */
 	u_short fatofs;		/* ch_cyl */
 #ifdef	HDDMOUNT
-	off64_t offset;
+	l_off_t offset;
 #endif
 	u_long fatsize;
 	u_short dirofs;
@@ -196,20 +206,20 @@ typedef struct _devstat {
 #define	ch_cyl	fatofs
 
 typedef struct _dosiobuf {
-	int _cnt;
+	long _cnt;
 	char *_ptr;
 	char *_base;
-	int _bufsize;
+	long _bufsize;
 	short _flag;
 	int _file;
 	long _top;
 	long _next;
 	long _off;
-	long _loc;
-	long _size;
+	off_t _loc;
+	off_t _size;
 	dent_t _dent;
 	long _clust;
-	u_short _offset;
+	long _offset;
 } dosFILE;
 
 #define	O_IOEOF	020000
@@ -243,6 +253,8 @@ typedef struct _dosdirdesc {
 	char *dd_buf;
 } dosDIR;
 
+#define	DID_IFDOSDRIVE	(-1)
+
 #ifdef	HDDMOUNT
 typedef struct _partition_t {
 	u_char boot __attribute__ ((packed));
@@ -303,9 +315,9 @@ typedef struct _partition98_t {
 #define	PT98_LINUX	0xe2	/* 0x80 | 0x62 */
 #endif	/* HDDMOUNT */
 
-#ifdef	NODNAMLEN
+#if	defined (DNAMESIZE) && DNAMESIZE < (MAXNAMLEN + 1)
 typedef struct _st_dirent {
-	char buf[sizeof(struct dirent) + MAXNAMLEN];
+	char buf[sizeof(struct dirent) - DNAMESIZE + MAXNAMLEN + 1];
 } st_dirent;
 #else
 typedef struct dirent	st_dirent;
@@ -326,7 +338,7 @@ extern VOID_T (*doswaitfunc)__P_((VOID_A));
 extern int (*dosintrfunc)__P_((VOID_A));
 
 #ifdef	HDDMOUNT
-extern off64_t *readpt __P_((char *, int));
+extern l_off_t *readpt __P_((char *, int));
 #endif
 extern int preparedrv __P_((int));
 extern int shutdrv __P_((int));
@@ -345,10 +357,10 @@ extern int dosstatfs __P_((int, char *));
 extern int dosstat __P_((char *, struct stat *));
 extern int doslstat __P_((char *, struct stat *));
 extern int dosaccess __P_((char *, int));
-#if	!MSDOS
+#ifndef	NOSYMLINK
 extern int dossymlink __P_((char *, char *));
 extern int dosreadlink __P_((char *, char *, int));
-#endif	/* !MSDOS */
+#endif
 extern int doschmod __P_((char *, int));
 #ifdef	USEUTIME
 extern int dosutime __P_((char *, struct utimbuf *));
@@ -361,7 +373,7 @@ extern int dosopen __P_((char *, int, int));
 extern int dosclose __P_((int));
 extern int dosread __P_((int, char *, int));
 extern int doswrite __P_((int, char *, int));
-extern off64_t doslseek __P_((int, off64_t, int));
+extern off_t doslseek __P_((int, off_t, int));
 extern int dosmkdir __P_((char *, int));
 extern int dosrmdir __P_((char *));
 extern int dosfileno __P_((FILE *));
