@@ -30,6 +30,8 @@
 #define	_NOTRADLAYOUT
 #define	_NOEXTRAWIN
 #define	_NOPTY
+#define	_NOEXTRAATTR
+#define	_NOLOGGING
 #endif	/* FD < 2 */
 
 #ifndef	__FD_PRIMAL__
@@ -41,8 +43,8 @@
 #endif
 
 #ifdef	DEBUG
-extern VOID mtrace __P_ ((VOID));
-extern VOID muntrace __P_ ((VOID));
+extern VOID mtrace __P_ ((VOID_A));
+extern VOID muntrace __P_ ((VOID_A));
 extern char *_mtrace_file;
 #endif
 
@@ -146,6 +148,15 @@ extern char *_mtrace_file;
 #endif
 #define	PROMPT2		"> "
 #define	DUMBSHELL	0
+#define	PTYMODE		0
+#define	PTYTERM		"vt100"
+#define	PTYMENUKEY	-1
+#define	LOGFILE		""
+#define	LOGSIZE		1024
+#define	USESYSLOG	0
+#define	LOGLEVEL	0
+#define	ROOTLOGLEVEL	1
+#define	THRUARGS	0
 #define	UNICODEBUFFER	0
 #define	SJISPATH	""
 #define	EUCPATH		""
@@ -160,9 +171,6 @@ extern char *_mtrace_file;
 #define	UTF8PATH	""
 #define	UTF8MACPATH	""
 #define	NOCONVPATH	""
-#define	PTYMODE		0
-#define	PTYTERM		"vt100"
-#define	PTYMENUKEY	-1
 
 
 /****************************************************************
@@ -187,6 +195,7 @@ extern char *_mtrace_file;
 #endif
 #define	MAXHISTNO	MAXTYPE(short)
 #define	MAXINVOKEARGS	MAXWINDOWS
+#define	MAXLOGLEN	255
 
 #ifdef	_NOSPLITWIN
 #undef	MAXWINDOWS
