@@ -59,7 +59,7 @@ extern int dosdrive;
 # else
 # define	L_SET	0
 # endif
-#endif
+#endif	/* !L_SET */
 
 #ifdef	_NODOSDRIVE
 #define	nodoschdir	Xchdir
@@ -207,15 +207,8 @@ CONST VOID_P vp2;
 
 	if (!isdir(namep1) && isdir(namep2)) return(1);
 	if (isdir(namep1) && !isdir(namep2)) return(-1);
-
-	if (namep1 -> name[0] == '.' && namep1 -> name[1] == '\0') return(-1);
-	if (namep2 -> name[0] == '.' && namep2 -> name[1] == '\0') return(1);
-	if (namep1 -> name[0] == '.' && namep1 -> name[1] == '.'
-	&& namep1 -> name[2] == '\0')
-		return(-1);
-	if (namep2 -> name[0] == '.' && namep2 -> name[1] == '.'
-	&& namep2 -> name[2] == '\0')
-		return(1);
+	if ((tmp = isdotdir(namep2 -> name) - isdotdir(namep1 -> name)))
+		return(tmp);
 
 	switch (sorton & 7) {
 		case 5:
@@ -295,10 +288,10 @@ CONST VOID_P vp2;
 		case 2:
 			if ((cp1 = strrchr(((treelist *)vp1) -> name, '.')))
 				cp1++;
-			else cp1 = "";
+			else cp1 = nullstr;
 			if ((cp2 = strrchr(((treelist *)vp2) -> name, '.')))
 				cp2++;
-			else cp2 = "";
+			else cp2 = nullstr;
 			tmp = strpathcmp2(cp1, cp2);
 			break;
 		case 3:
@@ -379,7 +372,7 @@ char *buf;
 #endif
 	else cp = underpath(cwd, homedir, len);
 
-	if (buf) strcpy(buf, (cp) ? cp : "");
+	if (buf) strcpy(buf, (cp) ? cp : nullstr);
 #ifdef	DEBUG
 	free(homedir);
 	homedir = NULL;
@@ -426,13 +419,56 @@ char *dir;
 	return(0);
 }
 
+#ifndef	NOFLOCK
+int lockfile(fd, mode)
+int fd, mode;
+{
+	static int lockmode[] = {
+# ifdef	USEFCNTLOCK
+		F_RDLCK, F_WRLCK, F_UNLCK,
+# else	/* !USEFCNTLOCK */
+#  ifdef	USELOCKF
+		F_LOCK, F_LOCK, F_ULOCK,
+#  else
+		LOCK_SH, LOCK_EX, LOCK_UN,
+#  endif
+# endif	/* !USEFCNTLOCK */
+	};
+# ifdef	USEFCNTLOCK
+	struct flock lock;
+# endif
+
+# ifndef	_NODOSDRIVE
+	if (fd >= DOSFDOFFSET) return(0);
+# endif
+
+# ifdef	USEFCNTLOCK
+	lock.l_type = lockmode[mode];
+	lock.l_start = lock.l_len = (off_t)0;
+	lock.l_whence = SEEK_SET;
+	return(fcntl(fd, F_SETLKW, &lock));
+# else	/* !USEFCNTLOCK */
+#  ifdef	USELOCKF
+	return(lockf(fd, lockmode[mode], (off_t)0));
+#  else
+	return(flock(fd, lockmode[mode]));
+#  endif
+# endif	/* !USEFCNTLOCK */
+}
+#endif	/* !NOFLOCK */
+
 int touchfile(path, stp)
 char *path;
 struct stat *stp;
 {
-	struct stat st, dummy;
+#ifndef	_NOEXTRAATTR
+	int i;
+#endif
+	struct stat st;
+	u_int mode;
 	int ret, duperrno;
 
+	if (!(stp -> st_nlink)) return(0);
 	if (Xlstat(path, &st) < 0) return(-1);
 	if (s_islnk(&st)) return(1);
 
@@ -443,10 +479,19 @@ struct stat *stp;
 #if	MSDOS
 	if (!(st.st_mode & S_IWRITE)) Xchmod(path, (st.st_mode | S_IWRITE));
 #endif
-	if (stp -> st_atime != (time_t)-1 && stp -> st_mtime != (time_t)-1) {
+
+	if (stp -> st_nlink & (TCH_ATIME | TCH_MTIME)) {
 #ifdef	USEUTIME
 		struct utimbuf times;
+#else
+		struct timeval tvp[2];
+#endif
 
+		if (!(stp -> st_nlink & TCH_ATIME))
+			stp -> st_atime = st.st_atime;
+		if (!(stp -> st_nlink & TCH_MTIME))
+			stp -> st_mtime = st.st_mtime;
+#ifdef	USEUTIME
 		times.actime = stp -> st_atime;
 		times.modtime = stp -> st_mtime;
 		if (Xutime(path, &times) < 0) {
@@ -454,8 +499,6 @@ struct stat *stp;
 			ret = -1;
 		}
 #else
-		struct timeval tvp[2];
-
 		tvp[0].tv_sec = stp -> st_atime;
 		tvp[0].tv_usec = 0;
 		tvp[1].tv_sec = stp -> st_mtime;
@@ -466,26 +509,36 @@ struct stat *stp;
 		}
 #endif
 	}
+
 #ifdef	HAVEFLAGS
-	if (stp -> st_flags != (u_long)-1) {
-# ifndef	_NODOSDRIVE
-		if (dospath(path, NULL)) {
-			duperrno = EACCES;
-			ret = -1;
-		}
-		else
-# endif
-		if (chflags(path, stp -> st_flags) < 0) {
+	if (stp -> st_nlink & TCH_FLAGS) {
+		if (Xchflags(path, stp -> st_flags) < 0) {
 			duperrno = errno;
 			ret = -1;
 		}
 	}
 #endif
-	dummy.st_mode = (u_short)-1;
-	if (stp -> st_mode != dummy.st_mode) {
-		stp -> st_mode &= ~S_IFMT;
-		stp -> st_mode |= (st.st_mode & S_IFMT);
-		if (Xchmod(path, stp -> st_mode) < 0) {
+
+	if (stp -> st_nlink & TCH_MODE) {
+		mode = stp -> st_mode;
+		mode &= ~S_IFMT;
+		mode |= (st.st_mode & S_IFMT);
+#ifndef	_NOEXTRAATTR
+		if ((stp -> st_nlink & TCH_MODEEXE)
+		&& !s_isdir(&st) && !(st.st_mode & S_IEXEC_ALL)) {
+			for (i = 0; i < 3; i++) {
+				if (!(stp -> st_mode & (1 << (i + 12))))
+					continue;
+				mode &= ~(1 << (i * 3));
+				mode |= (st.st_mode & (1 << (i * 3)));
+			}
+		}
+		if (stp -> st_nlink & TCH_MASK) {
+			mode &= ~(stp -> st_size);
+			mode |= (st.st_mode & (stp -> st_size));
+		}
+#endif	/* !_NOEXTRAATTR */
+		if (Xchmod(path, mode) < 0) {
 			duperrno = errno;
 			ret = -1;
 		}
@@ -493,14 +546,18 @@ struct stat *stp;
 #if	MSDOS
 	else if (!(st.st_mode & S_IWRITE)) Xchmod(path, st.st_mode);
 #endif
+
 #ifndef	NOUID
-	if (stp -> st_gid != (gid_t)-1) {
-# ifndef	_NODOSDRIVE
-		if (dospath(path, NULL)) /*EMPTY*/;
-		else
-# endif
-		if (chown(path, stp -> st_uid, stp -> st_gid) < 0)
-			chown(path, (uid_t)-1, stp -> st_gid);
+	if (stp -> st_nlink & (TCH_UID | TCH_GID)) {
+		if (!(stp -> st_nlink & TCH_UID)) stp -> st_uid = (uid_t)-1;
+		if (!(stp -> st_nlink & TCH_GID)) stp -> st_gid = (gid_t)-1;
+		if (Xchown(path, stp -> st_uid, stp -> st_gid) >= 0) /*EMPTY*/;
+		else if (!(stp -> st_nlink & TCH_CHANGE))
+			Xchown(path, (uid_t)-1, stp -> st_gid);
+		else {
+			duperrno = errno;
+			ret = -1;
+		}
 	}
 #endif	/* !NOUID */
 	if (ret < 0) errno = duperrno;
@@ -591,13 +648,13 @@ char *path;
 		cp = NOCWD_K;
 	else if ((cp = gethomedir()) && !nodoschdir(cp) && nodosgetwd(path))
 		cp = GOHOM_K;
-	else if (!nodoschdir(_SS_) && nodosgetwd(path)) cp = GOROT_K;
-	else error(_SS_);
+	else if (!nodoschdir(rootpath) && nodosgetwd(path)) cp = GOROT_K;
+	else error(rootpath);
 
 	if (path != fullpath) strncpy2(fullpath, path, MAXPATHLEN - 1);
 	warning(0, cp);
 #ifndef	_NOUSEHASH
-	searchhash(NULL, "", "");
+	searchhash(NULL, nullstr, nullstr);
 #endif
 	errno = duperrno;
 }
@@ -668,13 +725,13 @@ struct stat *stp1, *stp2;
 	if (issamebody(src, dest)) {
 		flags |= O_EXCL;
 		if (Xunlink(dest) < 0) {
-			Xclose(fd1);
+			VOID_C Xclose(fd1);
 			return(-1);
 		}
 	}
 #endif
 	if ((fd2 = Xopen(dest, flags, mode)) < 0) {
-		Xclose(fd1);
+		VOID_C Xclose(fd1);
 		return(-1);
 	}
 
@@ -694,17 +751,15 @@ struct stat *stp1, *stp2;
 		if (i < BUFSIZ) break;
 	}
 
-	Xclose(fd2);
-	Xclose(fd1);
+	VOID_C Xclose(fd2);
+	VOID_C Xclose(fd1);
 
 	if (i < 0) {
 		Xunlink(dest);
 		errno = duperrno;
 		return(-1);
 	}
-#ifdef	HAVEFLAGS
-	stp1 -> st_flags = (u_long)-1;
-#endif
+	stp1 -> st_nlink = (TCH_ATIME | TCH_MTIME);
 #ifdef	_USEDOSCOPY
 	if (touchfile(dest, stp1) < 0) return(-1);
 #else
@@ -759,10 +814,9 @@ struct stat *stp1, *stp2;
 	if (errno != EXDEV || s_isdir(stp1)) return(-1);
 	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
 		return(-1);
-#ifdef	HAVEFLAGS
-	stp1 -> st_flags = (u_long)-1;
-#endif
 
+	stp1 -> st_nlink = (TCH_MODE | TCH_UID | TCH_GID
+		| TCH_ATIME | TCH_MTIME);
 	return (touchfile(dest, stp1));
 }
 
@@ -1025,7 +1079,7 @@ char *dir, *file;
 	strcatdelim2(buf, dir, file);
 
 	if (rawchdir(buf) != 0) return(0);
-	rawchdir(_SS_);
+	rawchdir(rootpath);
 #if	MSDOS
 	spawnlpe(P_WAIT, "DELTREE.EXE", "DELTREE", "/Y", buf, NULL, environ);
 #else
@@ -1259,7 +1313,7 @@ char *tmpdir, *old;
 					fd = Xopen(fname,
 						O_CREAT | O_EXCL, 0600);
 					if (fd >= 0) {
-						Xclose(fd);
+						VOID_C Xclose(fd);
 						return(fname);
 					}
 					if (errno != EEXIST) break;
@@ -1313,7 +1367,7 @@ off_t bsiz;
 		tmp[i] = ch + 1;
 	}
 	tmp[i] = 0;
-	Xclose(fd);
+	VOID_C Xclose(fd);
 
 	return(tmp);
 }
@@ -1454,7 +1508,7 @@ int fs;
 	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	waitmes();
 
-	if (!(dirp = Xopendir("."))) {
+	if (!(dirp = Xopendir(curpath))) {
 		freevar(fnamelist);
 		noconv--;
 		lostcwd(path);
@@ -1504,8 +1558,8 @@ int fs;
 
 #if	!MSDOS
 	if (fs != FSID_EFS) entnum = NULL;	/* except IRIX File System */
-	else if (!(entnum = getentnum(".", persec))) {
-		warning(-1, ".");
+	else if (!(entnum = getentnum(curpath, persec))) {
+		warning(-1, curpath);
 		restorefile(tmpdir, path, fnamp);
 		freevar(fnamelist);
 		noconv--;
@@ -1513,8 +1567,8 @@ int fs;
 	}
 	totalent = headbyte
 		+ realdirsiz(tmpdir, dos, boundary, dirsize, namofs)
-		+ realdirsiz(".", dos, boundary, dirsize, namofs)
-		+ realdirsiz("..", dos, boundary, dirsize, namofs);
+		+ realdirsiz(curpath, dos, boundary, dirsize, namofs)
+		+ realdirsiz(parentpath, dos, boundary, dirsize, namofs);
 	block = tmpno = 0;
 	ptr = 3;
 	totalptr = 0;
@@ -1579,7 +1633,8 @@ int fs;
 		free(tmpdir);
 		tmpdir = cp;
 	}
-	fnamp = strcatdelim2(path, tmpdir, fnamelist[top]) - path;
+	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
+	snprintf2(&(path[fnamp]), sizeof(path) - fnamp, fnamelist[top]);
 	if (saferename(path, fnamelist[top]) < 0) warning(-1, path);
 	restorefile(tmpdir, path, fnamp);
 
