@@ -33,6 +33,7 @@ extern int internal_status;
 #ifdef	SIGALRM
 extern int noalrm;
 #endif
+extern int fdmode;
 extern int fdflags;
 extern int wheader;
 extern int ptymode;
@@ -43,6 +44,7 @@ extern int parentfd;
 extern char *ptytmpfile;
 
 static int (*lastfunc)__P_((VOID_A)) = NULL;
+static u_long lockflags = 0;
 
 static int waitpty __P_((VOID_A));
 static int NEAR ptygetkey __P_((VOID_A));
@@ -99,7 +101,7 @@ int isnl;
 	int dupdumbterm;
 
 	if (!emupid) {
-		syncptyout();
+		syncptyout(-1, -1);
 		VOID_C ttyiomode(isnl);
 	}
 	else {
@@ -116,7 +118,7 @@ VOID Xstdiomode(VOID_A)
 	int dupdumbterm;
 
 	if (!emupid) {
-		syncptyout();
+		syncptyout(-1, -1);
 		VOID_C stdiomode();
 	}
 	else {
@@ -134,8 +136,10 @@ int init;
 	int mode, dupdumbterm;
 
 	dupdumbterm = dumbterm;
-	dumbterm = 1;
-	if (ptymode) dumbterm = 1;
+#ifndef	_NOORIGSHELL
+	if (isshptymode()) dumbterm = 1;
+#endif
+	if (isptymode()) dumbterm = 1;
 	mode = termmode(init);
 	dumbterm = dupdumbterm;
 
@@ -232,10 +236,10 @@ va_dcl
 	int n;
 
 	VA_START(args, fmt);
-
 	n = vasprintf2(&buf, fmt, args);
 	va_end(args);
 	if (n < 0) error("malloc()");
+
 	if (!emupid) VOID_C cputs2(buf);
 	else {
 		sendword(emufd, TE_CPUTS2);
@@ -358,8 +362,29 @@ VOID changekcode(VOID_A)
 	if (!emupid) return;
 
 	sendword(emufd, TE_CHANGEKCODE);
+	sendword(emufd, inputkcode);
 	sendword(emufd, outputkcode);
 }
+
+VOID changeinkcode(VOID_A)
+{
+	ptylist[win].incode = ptyinkcode;
+	if (!emupid) return;
+
+	sendword(emufd, TE_CHANGEINKCODE);
+	sendword(emufd, win);
+	sendword(emufd, ptyinkcode);
+}
+
+VOID changeoutkcode(VOID_A)
+{
+	ptylist[win].outcode = ptyoutkcode;
+	if (!emupid) return;
+
+	sendword(emufd, TE_CHANGEOUTKCODE);
+	sendword(emufd, win);
+	sendword(emufd, ptyoutkcode);
+}
 #endif	/* !_NOKANJICONV */
 
 static int NEAR ptygetkey(VOID_A)
@@ -370,6 +395,13 @@ static int NEAR ptygetkey(VOID_A)
 	for (;;) {
 		n = -1;
 		c = getkey2(sigalrm());
+		while (lockflags & (1 << win)) {
+			kbhit2(1000000L / SENSEPERSEC);
+			waitpty();
+		}
+#ifndef	_NOORIGSHELL
+		if (isshptymode()) break;
+#endif
 		if (c < 0 || ptymenukey < 0 || c != ptymenukey) break;
 
 		str[0] = asprintf3(PTYAI_K, getkeysym(ptymenukey, 0));
@@ -804,6 +836,35 @@ int w;
 			deletedrv(n);
 			break;
 #endif	/* _USEDOSEMU */
+		case TE_LOCKFRONT:
+			lockflags |= (1 << w);
+			if (!emupid) break;
+			sendword(emufd, TE_LOCKBACK);
+			sendword(emufd, w);
+			break;
+		case TE_UNLOCKFRONT:
+			lockflags &= ~(1 << w);
+			if (!emupid) break;
+			sendword(emufd, TE_UNLOCKBACK);
+			sendword(emufd, w);
+			break;
+		case TE_SAVETTYIO:
+			if (recvbuf(fd, &n, sizeof(n)) < 0
+			|| recvbuf(fd, &val, sizeof(val)) < 0) break;
+			if (!val) cp = NULL;
+			else {
+				cp = malloc2(val);
+				if (recvbuf(fd, cp, val) < 0) {
+					free(cp);
+					break;
+				}
+			}
+			if (n >= 0) {
+				if (duptty[n]) free(duptty[n]);
+				duptty[n] = cp;
+			}
+			else if (cp) free(cp);
+			break;
 		case TE_CHANGESTATUS:
 			if (recvbuf(fd, &n, sizeof(n)) < 0) break;
 			ptylist[w].status = n;
@@ -830,7 +891,7 @@ int frontend(VOID_A)
 
 	if (ch < -1) {
 		status = 0;
-		internal_status = 1;
+		internal_status = FNC_CANCEL;
 	}
 	else {
 		status = ptylist[win].status;
@@ -839,7 +900,7 @@ int frontend(VOID_A)
 			recvchild(win);
 			killpty(win, &status);
 		}
-		internal_status = (status < 128) ? status : -2;
+		internal_status = (status < 128) ? status : FNC_FAIL;
 	}
 
 	sigvecset(n);
