@@ -77,6 +77,19 @@ extern char *sys_errlist[];
 #define	strerror2(n)		(char *)sys_errlist[n]
 #endif
 
+#ifdef	_NOPTY
+#define	isptymode()		0
+#define	isshptymode()		0
+#else	/* !_NOPTY */
+# ifdef	_NOORIGSHELL
+# define	isptymode()	ptymode
+# define	isshptymode()	0
+# else
+# define	isptymode()	(ptymode && (fdmode || !shptymode))
+# define	isshptymode()	(!fdmode && shptymode)
+# endif
+#endif	/* !_NOPTY */
+
 /* main.c */
 extern VOID error __P_((char *));
 extern VOID setlinecol __P_((VOID_A));
@@ -90,6 +103,8 @@ extern VOID title __P_((VOID_A));
 extern VOID saveorigenviron __P_((VOID_A));
 #endif
 extern int loadruncom __P_((char *, int));
+extern VOID initfd __P_((char **));
+extern VOID prepareexitfd __P_((int));
 
 /* system.c */
 #ifdef	_NOORIGSHELL
@@ -102,6 +117,23 @@ extern FILE *dopopen __P_((char *));
 extern int dopclose __P_((FILE *));
 #endif
 
+/* log.c */
+#ifdef	_NOLOGGING
+#define	LOG0(l, n, f)
+#define	LOG1(l, n, f, a1)
+#define	LOG2(l, n, f, a1, a2)
+#define	LOG3(l, n, f, a1, a2, a3)
+#else	/* !_NOLOGGING */
+#define	LOG0(l, n, f)		logsyscall(l, n, f)
+#define	LOG1(l, n, f, a1)	logsyscall(l, n, f, a1)
+#define	LOG2(l, n, f, a1, a2)	logsyscall(l, n, f, a1, a2)
+#define	LOG3(l, n, f, a1, a2, a3) \
+				logsyscall(l, n, f, a1, a2, a3)
+extern VOID logclose __P_((VOID_A));
+extern VOID logsyscall __P_((int, int, CONST char *, ...));
+extern VOID logmessage __P_((int, CONST char *, ...));
+#endif	/* !_NOLOGGING */
+
 /* dosemu.c or unixemu.c */
 #ifdef	_USEDOSPATH
 extern int _dospath __P_((char *));
@@ -123,7 +155,7 @@ extern VOID Xrewinddir __P_((DIR *));
 #if	MSDOS
 extern int rawchdir __P_((char *));
 #else
-#define	rawchdir	chdir
+#define	rawchdir(p)	((chdir(p)) ? -1 : 0)
 #endif
 extern int Xchdir __P_((char *));
 extern char *Xgetwd __P_((char *));
@@ -138,11 +170,17 @@ extern int Xutime __P_((char *, struct utimbuf *));
 #else
 extern int Xutimes __P_((char *, struct timeval []));
 #endif
+#ifdef	HAVEFLAGS
+extern int Xchflags __P_((char *, u_long));
+#endif
+#ifndef	NOUID
+extern int Xchown __P_((char *, uid_t, gid_t));
+#endif
 extern int Xunlink __P_((char *));
 extern int Xrename __P_((char *, char *));
 extern int Xopen __P_((char *, int, int));
 #ifdef	_NODOSDRIVE
-#define	Xclose		close
+#define	Xclose(f)	((close(f)) ? -1 : 0)
 #define	Xread		read
 #define	Xwrite		write
 #define	Xlseek		lseek
@@ -165,6 +203,7 @@ extern FILE *Xfopen __P_((char *, char *));
 #ifdef	_NODOSDRIVE
 #define	Xfdopen		fdopen
 #define	Xfclose		fclose
+#define	Xfileno		fileno
 #define	Xfeof		feof
 #define	Xfread		fread
 #define	Xfwrite		fwrite
@@ -176,6 +215,7 @@ extern FILE *Xfopen __P_((char *, char *));
 #else
 extern FILE *Xfdopen __P_((int, char *));
 extern int Xfclose __P_((FILE *));
+extern int Xfileno __P_((FILE *));
 extern int Xfeof __P_((FILE *));
 extern int Xfread __P_((char *, int, int, FILE *));
 extern int Xfwrite __P_((char *, int, int, FILE *));
@@ -210,7 +250,7 @@ extern p_id_t Xforkpty __P_((int *, char *, char *));
 #else	/* !_NOPTY */
 extern VOID regionscroll __P_((int, int, int, int, int, int));
 extern int selectpty __P_((int, int [], char [], int));
-extern VOID syncptyout __P_((VOID_A));
+extern VOID syncptyout __P_((int, int));
 extern int recvbuf __P_((int, VOID_P, int));
 extern VOID sendbuf __P_((int, VOID_P, int));
 extern int recvword __P_((int, int *));
@@ -269,6 +309,8 @@ extern VOID insertwin __P_((int, int));
 extern VOID deletewin __P_((int, int));
 # ifndef	_NOKANJICONV
 extern VOID changekcode __P_((VOID_A));
+extern VOID changeinkcode __P_((VOID_A));
+extern VOID changeoutkcode __P_((VOID_A));
 # endif
 extern int frontend __P_((VOID_A));
 #endif	/* !_NOPTY */
@@ -336,6 +378,11 @@ extern int cmptree __P_((CONST VOID_P, CONST VOID_P));
 extern struct dirent *searchdir __P_((DIR *, reg_t *, char *));
 extern int underhome __P_((char *));
 extern int preparedir __P_((char *));
+#ifdef	NOFLOCK
+#define	lockfile(f, m)
+#else
+extern int lockfile __P_((int, int));
+#endif
 extern int touchfile __P_((char *, struct stat *));
 #ifdef	_NODOSDRIVE
 #define	nodoslstat	Xlstat
@@ -453,6 +500,9 @@ extern int parsesetdrv __P_((int, char *[], devinfo *));
 extern int parsekeymap __P_((int, char *[], keyseq_t *));
 extern VOID printkeymap __P_((keyseq_t *, int, FILE *));
 #endif
+#if	!MSDOS
+extern int savestdio __P_((int));
+#endif
 #ifdef	_NOORIGSHELL
 extern int searchalias __P_((char *, int));
 extern int addalias __P_((char *, char *));
@@ -500,6 +550,7 @@ extern VOID freedefine __P_((VOID_A));
 #define	BL_KCONV	"kconv"
 #define	BL_READLINE	"readline"
 #define	BL_YESNO	"yesno"
+#define	BL_SAVETTY	"savetty"
 #define	BL_ALIAS	"alias"
 #define	BL_UALIAS	"unalias"
 #define	BL_FUNCTION	"function"
@@ -565,6 +616,9 @@ extern char *newkanjiconv __P_((char *, int, int, int));
 #ifndef	_NOKANJIFCONV
 extern int getkcode __P_((char *));
 #endif
+#ifndef	_NOKANJICONV
+extern int getoutputkcode __P_((VOID_A));
+#endif
 extern char *convget __P_((char *, char *, int));
 extern char *convput __P_((char *, char *, int, int, char *, int *));
 
@@ -643,6 +697,7 @@ extern int customize __P_((VOID_A));
 #endif	/* !_NOCUSTOMIZE */
 
 /* command.c */
+extern int evalstatus __P_((int));
 #ifndef	_NOSPLITWIN
 extern int nextwin __P_((VOID_A));
 #endif
@@ -653,6 +708,10 @@ extern int putmode __P_((char *, u_int, int));
 #ifdef	HAVEFLAGS
 extern int putflags __P_((char *, u_long));
 #endif
+#ifndef	NOUID
+extern int putowner __P_((char *, uid_t));
+extern int putgroup __P_((char *, gid_t));
+#endif
 extern VOID waitmes __P_((VOID_A));
 extern int filetop __P_((int));
 extern int calcwidth __P_((VOID_A));
@@ -665,4 +724,4 @@ extern VOID calcwin __P_((VOID_A));
 extern VOID movepos __P_((int, int));
 extern VOID rewritefile __P_((int));
 extern VOID addlist __P_((VOID_A));
-extern VOID main_fd __P_((char **));
+extern VOID main_fd __P_((char **, int));
