@@ -26,6 +26,8 @@
 #define	NOTZFILEH
 #define	USETIMEH
 #define	USEUTIME
+#define	NOFLOCK
+#define	NOSYSLOG
 #define	USEMKTIME
 #define	USESTRERROR
 #define	SENSEPERSEC	20
@@ -594,8 +596,6 @@ typedef long	off_t;
 #define	DECLSIGLIST
 #define	DECLERRLIST
 #define	USELEAPCNT
-#define	USEMOUNTH
-#define	USEMNTINFO
 #define	USEFFSTYPE
 #define	USERE_COMP
 #define	USESETENV
@@ -605,6 +605,13 @@ typedef long	off_t;
 # if	defined (NetBSD1_0) && (NetBSD1_0 < 1)
 # define	USEFFSIZE
 # endif
+# if	defined (__NetBSD_Version__) && (__NetBSD_Version__ >= 300000000)
+# define	USESTATVFSH
+# define	USEGETVFSTAT
+# else
+# define	USEMOUNTH
+# define	USEMNTINFO
+# endif
 #endif
 
 #if	defined (__bsdi__)
@@ -828,6 +835,7 @@ typedef long	off_t;
 /* #define USESTDARGH	;use <stdarg.h> for va_list */
 /* #define USEMKDEVH	;use <sys/mkdev.h> for major()/minor() */
 /* #define USEMKNODH	;use <sys/mknod.h> for major()/minor() */
+/* #define USELOCKFH	;use <sys/lockf.h> for lockf() */
 /* #define USESGTTY	;use sgtty interface */
 /* #define USETERMIO	;use termio interface */
 /* #define USETERMIOS	;use termios interface */
@@ -856,6 +864,7 @@ typedef long	off_t;
 /* #define USEMNTENTH	;use <mntent.h> as header of the mount entry */
 /* #define USEMNTTABH	;use <sys/mnttab.h> as header of the mount entry */
 /* #define USEGETFSSTAT	;use getfsstat() to get the mount entry */
+/* #define USEGETVFSTAT	;use getvfsstat() to get the mount entry */
 /* #define USEMNTCTL	;use mntctl() to get the mount entry */
 /* #define USEMNTINFOR	;use getmntinfo_r() to get the mount entry */
 /* #define USEMNTINFO	;use getmntinfo() to get the mount entry */
@@ -880,6 +889,10 @@ typedef long	off_t;
 /* #define NOTERMVAR	;have not termcap variables such as PC, ospeed, etc. */
 /* #define USEUTIME	;use utime() instead of utimes() */
 /* #define USEGETWD	;use getwd() instead of getcwd() */
+/* #define USEFCNTLOCK	;use fcntl() lock instead of flock() */
+/* #define USELOCKF	;use lockf() instead of flock() */
+/* #define NOFLOCK	;have neither flock() nor lockf() */
+/* #define NOSYSLOG	;have not syslog() */
 /* #define USETIMELOCAL	;have timelocal() as inverse of localtime() */
 /* #define USEMKTIME	;use mktime() instead of timelocal() */
 /* #define USESYSCONF	;use sysconf() for getting system configuration */
@@ -971,6 +984,10 @@ typedef long	off_t;
 #define	USEDEVPTMX
 #endif
 
+#if	defined (POSIX) || defined (SYSV)
+#define	USEFCNTLOCK
+#endif
+
 /*                                 */
 /* Eval configurations by Configur */
 /*                                 */
@@ -1119,41 +1136,51 @@ typedef long	off_t;
 #endif
 
 #if	defined (USEMNTTABH) || defined (USEGETFSSTAT)
+# ifdef	USEGETVFSTAT
+# undef	USEGETVFSTAT
+# endif
+#endif
+
+#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) \
+|| defined (USEGETVFSTAT)
 # ifdef	USEMNTCTL
 # undef	USEMNTCTL
 # endif
 #endif
 
-#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) || defined (USEMNTCTL)
+#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) \
+|| defined (USEGETVFSTAT) || defined (USEMNTCTL)
 # ifdef	USEMNTINFOR
 # undef	USEMNTINFOR
 # endif
 #endif
 
-#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) || defined (USEMNTCTL) \
-|| defined (USEMNTINFOR)
+#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) \
+|| defined (USEGETVFSTAT) || defined (USEMNTCTL) || defined (USEMNTINFOR)
 # ifdef	USEMNTINFO
 # undef	USEMNTINFO
 # endif
 #endif
 
-#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) || defined (USEMNTCTL) \
-|| defined (USEMNTINFOR) || defined (USEMNTINFO)
+#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) \
+|| defined (USEGETVFSTAT) || defined (USEMNTCTL) || defined (USEMNTINFOR) \
+|| defined (USEMNTINFO)
 # ifdef	USEGETMNT
 # undef	USEGETMNT
 # endif
 #endif
 
-#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) || defined (USEMNTCTL) \
-|| defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETMNT)
+#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) \
+|| defined (USEGETVFSTAT) || defined (USEMNTCTL) || defined (USEMNTINFOR) \
+|| defined (USEMNTINFO) || defined (USEGETMNT)
 # ifdef	USEGETFSENT
 # undef	USEGETFSENT
 # endif
 #endif
 
-#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) || defined (USEMNTCTL) \
-|| defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETMNT) \
-|| defined (USEGETFSENT)
+#if	defined (USEMNTTABH) || defined (USEGETFSSTAT) \
+|| defined (USEGETVFSTAT) || defined (USEMNTCTL) || defined (USEMNTINFOR) \
+|| defined (USEMNTINFO) || defined (USEGETMNT) || defined (USEGETFSENT)
 # ifdef	USEMNTENTH
 # undef	USEMNTENTH
 # endif
