@@ -59,6 +59,7 @@ extern int maxarchive;
 extern int wheader;
 extern char fullpath[];
 extern char *histfile;
+extern int savehist;
 extern char *helpindex[];
 extern int subwindow;
 extern int win_x;
@@ -72,8 +73,9 @@ extern char *unitblpath;
 #endif
 #ifndef	_NOPTY
 extern int ptymode;
+extern int ptyinternal;
 extern int parentfd;
-#endif	/* !_NOPTY */
+#endif
 
 #define	CLOCKUPDATE	10	/* sec */
 
@@ -146,6 +148,11 @@ static VOID NEAR printtime __P_((int));
 #ifdef	SIGALRM
 static int trapalrm __P_((VOID_A));
 #endif
+static char *NEAR getversion __P_((int *));
+#ifndef	_NOLOGGING
+static VOID NEAR startlog __P_((char **));
+static VOID NEAR endlog __P_((int));
+#endif
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline __P_((char *, char *, int, char *));
 #endif
@@ -154,7 +161,6 @@ static int NEAR evaloption __P_((char *[]));
 static char *NEAR searchenv __P_((char *, char *[]));
 static VOID NEAR setexecname __P_((char *));
 static VOID NEAR setexecpath __P_((char *, char *[]));
-static VOID NEAR prepareexitfd __P_((VOID_A));
 
 char *origpath = NULL;
 char *progpath = NULL;
@@ -164,6 +170,9 @@ char *tmpfilename = NULL;
 int adjtty = 0;
 #endif
 int showsecond = 0;
+#if	FD >= 2
+int thruargs = 0;
+#endif
 int hideclock = 0;
 #ifdef	SIGALRM
 int noalrm = 1;
@@ -235,7 +244,7 @@ char *s;
 	if (isorgpid()) {
 		inittty(1);
 		keyflush();
-		prepareexitfd();
+		prepareexitfd(2);
 #ifndef	_NOORIGSHELL
 # ifndef	NOJOB
 		killjob();
@@ -266,6 +275,9 @@ int sig;
 		forcecleandir(deftmpdir, tmpfilename);
 #ifndef	_NODOSDRIVE
 		dosallclose();
+#endif
+#ifndef	_NOLOGGING
+		endlog(sig + 128);
 #endif
 		endterm();
 		inittty(1);
@@ -659,7 +671,8 @@ int hide;
 	}
 	if (timersec-- < CLOCKUPDATE && !showsecond) return;
 #ifndef	_NOPTY
-	if (ptymode && hide <= 1) hide = 0;
+	if (checkallpty() < 0) rewritefile(0);
+	if (isptymode() && !ptyinternal && hide <= 1) hide = 0;
 #endif
 	if (hide) return;
 
@@ -695,6 +708,12 @@ static int trapalrm(VOID_A)
 
 	duperrno = errno;
 	signal2(SIGALRM, SIG_IGN);
+# if	!defined (_NOORIGSHELL) && !defined (_NOPTY)
+	if (isshptymode()) {
+		if (mypid == shellpid) printtime(hideclock);
+	}
+	else
+# endif
 	if (isorgpid()) printtime(hideclock);
 # ifdef	SIGWINCH
 	pollscreen(0);
@@ -752,9 +771,24 @@ int set;
 	return(old);
 }
 
-VOID title(VOID_A)
+static char *NEAR getversion(lenp)
+int *lenp;
 {
 	char *cp, *eol;
+
+	cp = strchr(version, ' ');
+	while (*(++cp) == ' ');
+	if (lenp) {
+		if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
+		*lenp = eol - cp;
+	}
+
+	return(cp);
+}
+
+VOID title(VOID_A)
+{
+	char *cp;
 	int i, len;
 
 	Xlocate(0, L_TITLE);
@@ -772,16 +806,13 @@ VOID title(VOID_A)
 	}
 	Xcputs2(" Ver.");
 	len += 5;
-	cp = strchr(version, ' ');
-	while (*(++cp) == ' ');
-	if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
-	i = eol - cp;
-	Xcprintf2("%-*.*s", i, i, cp);
+	cp = getversion(&i);
+	cputstr(i, cp);
 	if (distributor) {
 		Xputch2('#');
 		i++;
 	}
-	cp = (iswellomit()) ? "" : " (c)1995-2005 T.Shirai  ";
+	cp = (iswellomit()) ? nullstr : " (c)1995-2006 T.Shirai  ";
 	Xcputs2(cp);
 	i = n_column - len - strlen2(cp) - i;
 	while (i-- > 0) Xputch2(' ');
@@ -810,6 +841,61 @@ VOID saveorigenviron(VOID_A)
 }
 #endif	/* !_NOCUSTOMIZE */
 
+#ifndef	_NOLOGGING
+static VOID NEAR startlog(argv)
+char **argv;
+{
+# ifndef	NOUID
+	uid_t uid;
+	int lvl;
+# endif
+	char *cp, *tmp, buf[MAXLOGLEN + 1];
+	int i, len;
+
+	cp = buf;
+	for (i = 1; argv && argv[i]; i++) {
+		if (cp > buf) *(cp++) = ' ';
+		tmp = killmeta(argv[i]);
+		len = snprintf2(cp, sizeof(buf) - (cp - buf), "%s", tmp);
+		free(tmp);
+		if (len < 0) break;
+		cp += len;
+	}
+	*cp = '\0';
+
+	cp = getversion(&len);
+# ifdef	NOUID
+	logmessage(_LOG_DEBUG_, "%s (%-*.*s) starts; PWD=%k; ARGS=%k",
+		progname, len, len, cp, origpath, buf);
+# else
+	lvl = ((uid = getuid())) ? _LOG_DEBUG_ : _LOG_WARNING_;
+	logmessage(lvl, "%s (%-*.*s) starts; UID=%d; PWD=%k; ARGS=%k",
+		progname, len, len, cp, uid, origpath, buf);
+# endif
+}
+
+static VOID NEAR endlog(status)
+int status;
+{
+# ifndef	NOUID
+	uid_t uid;
+	int lvl;
+# endif
+	char cwd[MAXPATHLEN];
+
+	if (!Xgetwd(cwd)) strcpy(cwd, "?");
+# ifdef	NOUID
+	logmessage(_LOG_DEBUG_, "%s ends; PWD=%k; STATUS=%d",
+		progname, cwd, status);
+# else
+	lvl = ((uid = getuid())) ? _LOG_DEBUG_ : _LOG_WARNING_;
+	logmessage(lvl, "%s ends; UID=%d; PWD=%k; STATUS=%d",
+		progname, uid, cwd, status);
+# endif
+	logclose();
+}
+#endif	/* !_NOLOGGING */
+
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline(command, file, n, line)
 char *command, *file;
@@ -1049,15 +1135,42 @@ char *argv, *envp[];
 	}
 }
 
-static VOID NEAR prepareexitfd(VOID_A)
+/*ARGSUSED*/
+VOID initfd(argv)
+char **argv;
+{
+	if (interactive) {
+#if	!MSDOS
+		if (adjtty) {
+			Xstdiomode();
+			inittty(0);
+			Xttyiomode(0);
+		}
+#endif	/* !MSDOS */
+		loadhistory(0, histfile);
+		entryhist(1, origpath, 1);
+	}
+#ifndef	_NOLOGGING
+	startlog(argv);
+#endif
+}
+
+/*ARGSUSED*/
+VOID prepareexitfd(status)
+int status;
 {
 	char cwd[MAXPATHLEN];
 
+#ifndef	_NOLOGGING
+	endlog(status);
+#endif
+	if (interactive && savehist > 0) savehistory(0, histfile);
+
 	cwd[0] = '\0';
 	if (origpath && _chdir2(origpath) < 0) {
 		perror2(origpath);
 		if (!Xgetwd(cwd)) *cwd = '\0';
-		rawchdir(_SS_);
+		rawchdir(rootpath);
 	}
 	free(origpath);
 	free(progname);
@@ -1254,7 +1367,9 @@ char *argv[], *envp[];
 	cp = getshellname(progname, NULL, NULL);
 	if (!strpathcmp(cp, FDSHELL) || !strpathcmp(cp, "su")) {
 		i = main_shell(argc, argv, envp);
-		prepareexitfd();
+# ifndef	_NOPTY
+		killallpty();
+# endif
 		Xexit2(i);
 	}
 	fdmode = interactive = 1;
@@ -1294,31 +1409,26 @@ char *argv[], *envp[];
 
 	i = evaloption(argv);
 	checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
-#if	!MSDOS
-	if (adjtty) {
-		Xstdiomode();
-		inittty(0);
-		Xttyiomode(0);
-	}
-#endif	/* !MSDOS */
-
-	loadhistory(0, histfile);
-	entryhist(1, origpath, 1);
+	initfd(argv);
 	Xputterms(T_CLEAR);
 
 #ifdef	SIGWINCH
 	nowinch = 0;
 #endif
-	main_fd(&(argv[i]));
+#if	FD >= 2
+	main_fd(&(argv[i]), (thruargs) ? 1 : 0);
+#else
+	main_fd(&(argv[i]), 0);
+#endif
 	sigvecset(0);
-	prepareexitfd();
+#ifndef	_NOLOGGING
+	endlog(0);
+#endif
 
 	Xstdiomode();
-#ifndef	_NOORIGSHELL
-# ifndef	NOJOB
+#if	!defined (_NOORIGSHELL) && !defined (NOJOB)
 	killjob();
-# endif
-#endif	/* !_NOORIGSHELL */
+#endif
 #ifndef	_NOPTY
 	killallpty();
 #endif
