@@ -38,6 +38,7 @@ typedef off_t		l_off_t;
 #ifdef	FD
 # ifndef	_NOKANJICONV
 extern char *newkanjiconv __P_((char *, int, int, int));
+extern int getoutputkcode __P_((VOID_A));
 # endif
 extern char *restorearg __P_((char *));
 #endif
@@ -338,7 +339,8 @@ int width, prec;
 			}
 			tmp[j] = '\0';
 
-			cp = newkanjiconv(tmp, DEFCODE, outputkcode, L_OUTPUT);
+			cp = newkanjiconv(tmp, DEFCODE,
+				getoutputkcode(), L_OUTPUT);
 			if ((pbufp -> flags & (VF_NEW | VF_FILE))
 			|| pbufp -> ptr + strlen(cp) < pbufp -> size)
 				break;
@@ -359,7 +361,7 @@ int width, prec;
 		}
 	}
 	else
-#endif	/* !FD || _NOKANJICONV */
+#endif	/* FD && !_NOKANJICONV */
 	for (i = j = 0; i < len; i++, j++) {
 		c = s[j];
 		if (iskanji1(s, j)) {
@@ -548,7 +550,6 @@ va_list args;
 		if (len < 0) return(-1);
 		total += len;
 	}
-	va_end(args);
 	if (pbufp -> buf && !(pbufp -> flags & VF_FILE))
 		pbufp -> buf[pbufp -> ptr] = '\0';
 
@@ -586,13 +587,12 @@ va_dcl
 	printbuf_t pbuf;
 	int n;
 
-	VA_START(args, fmt);
-
 	if (!sp) return(-1);
 	pbuf.flags = VF_NEW;
+	VA_START(args, fmt);
 	n = commonprintf(&pbuf, fmt, args);
-	*sp = pbuf.buf;
 	va_end(args);
+	*sp = pbuf.buf;
 
 	return(n);
 }
@@ -629,12 +629,11 @@ va_dcl
 	printbuf_t pbuf;
 	int n;
 
-	VA_START(args, fmt);
-
 	if (size < 0) return(-1);
 	pbuf.buf = s;
 	pbuf.size = size;
 	pbuf.flags = 0;
+	VA_START(args, fmt);
 	n = commonprintf(&pbuf, fmt, args);
 	va_end(args);
 
@@ -656,10 +655,9 @@ va_dcl
 	printbuf_t pbuf;
 	int n;
 
-	VA_START(args, fmt);
-
 	pbuf.buf = (char *)fp;
 	pbuf.flags = VF_FILE;
+	VA_START(args, fmt);
 	n = commonprintf(&pbuf, fmt, args);
 	va_end(args);
 
