@@ -55,10 +55,6 @@ typedef struct _kstree_t {
 } kstree_t;
 #endif	/* !MSDOS */
 
-#ifdef	USESELECTH
-#include <sys/select.h>
-#endif
-
 #ifndef	NOUNISTDH
 #include <unistd.h>
 #endif
@@ -77,9 +73,7 @@ extern int errno;
 #include "termio.h"
 #include "term.h"
 
-#define	GETSIZE		"\033[6n"
-#define	SIZEFMT		"\033[%d;%dR"
-#define	WINTERMNAME	"iris"
+#define	WINTERMNAME		"iris"
 #ifdef	USESTRERROR
 #define	strerror2		strerror
 #else
@@ -90,7 +84,7 @@ extern char *sys_errlist[];
 #endif
 
 #ifdef	DEBUG
-extern VOID muntrace __P_ ((VOID));
+extern VOID muntrace __P_ ((VOID_A));
 extern char *_mtrace_file;
 #endif
 
@@ -340,16 +334,6 @@ extern int tputs __P_((char *, int, int (*)__P_((tputs_t))));
 #endif
 #endif	/* !MSDOS */
 
-#ifndef	FD_SET
-typedef struct fd_set {
-	u_int fds_bits[1];
-} fd_set;
-# define	FD_ZERO(p)	(((p) -> fds_bits[0]) = 0)
-# define	FD_SET(n, p)	(((p) -> fds_bits[0]) |= ((u_int)1 << (n)))
-#endif	/* !FD_SET */
-
-#define	MAXFDSET	256
-
 static int NEAR err2 __P_((char *));
 #if	!MSDOS
 static char *NEAR tstrdup __P_((char *));
@@ -413,6 +397,7 @@ int (*keywaitfunc)__P_((VOID_A)) = NULL;
 #if	!MSDOS
 int usegetcursor = 0;
 int suspended = 0;
+char *duptty[2] = {NULL, NULL};
 #endif
 int ttyio = -1;
 int isttyiomode = 0;
@@ -968,8 +953,7 @@ int cooked2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(0, CBREAK | RAW, LPASS8, LLITOUT | LPENDIN);
 #else
-	ttymode(ISIG | ICANON | IEXTEN, PENDIN,
-		BRKINT | IXON, IGNBRK | ISTRIP,
+	ttymode(TIO_LCOOKED, PENDIN, TIO_ICOOKED, ~TIO_INOCOOKED,
 		OPOST, 0, VAL_VMIN, VAL_VTIME);
 #endif
 
@@ -981,7 +965,7 @@ int cbreak2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(CBREAK, 0, LLITOUT, 0);
 #else
-	ttymode(ISIG | IEXTEN, ICANON, BRKINT | IXON, IGNBRK, OPOST, 0, 1, 0);
+	ttymode(TIO_LCBREAK, ICANON, TIO_ICOOKED, IGNBRK, OPOST, 0, 1, 0);
 #endif
 
 	return(0);
@@ -992,8 +976,7 @@ int raw2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(RAW, 0, LLITOUT, 0);
 #else
-	ttymode(0, ISIG | ICANON | IEXTEN,
-		IGNBRK, BRKINT | IXON, 0, OPOST, 1, 0);
+	ttymode(0, TIO_LCOOKED, IGNBRK, TIO_ICOOKED, 0, OPOST, 1, 0);
 #endif
 
 	return(0);
@@ -1004,7 +987,7 @@ int echo2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(ECHO, 0, LCRTBS | LCRTERA | LCRTKIL | LCTLECH, 0);
 #else
-	ttymode(ECHO | ECHOE | ECHOCTL | ECHOKE, ECHONL, 0, 0, 0, 0, 0, 0);
+	ttymode(TIO_LECHO, ECHONL, 0, 0, 0, 0, 0, 0);
 #endif
 
 	return(0);
@@ -1015,7 +998,7 @@ int noecho2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(0, ECHO, 0, LCRTBS | LCRTERA);
 #else
-	ttymode(0, ECHO | ECHOE | ECHOK | ECHONL, 0, 0, 0, 0, 0, 0);
+	ttymode(0, ~TIO_LNOECHO, 0, 0, 0, 0, 0, 0);
 #endif
 
 	return(0);
@@ -1026,7 +1009,7 @@ int nl2(VOID_A)
 #ifdef	USESGTTY
 	ttymode(CRMOD, 0, 0, 0);
 #else
-	ttymode(0, 0, ICRNL, 0, ONLCR, OCRNL | ONOCR | ONLRET, 0, 0);
+	ttymode(0, 0, ICRNL, 0, ONLCR, ~TIO_ONONL, 0, 0);
 #endif
 
 	return(0);
@@ -1081,28 +1064,54 @@ int keyflush(VOID_A)
 
 	return(0);
 }
+
+int savettyio(reset)
+int reset;
+{
+	char *tty;
+	int n;
+
+	if (reset) tty = NULL;
+	else {
+		savetermio(ttyio, &tty, NULL);
+		if (!tty) return(-1);
+	}
+
+	n = (isttyiomode) ? 0 : 1;
+	if (duptty[n]) free(duptty[n]);
+	duptty[n] = tty;
+
+	return(n);
+}
 #endif	/* !MSDOS */
 
 int ttyiomode(isnl)
 int isnl;
 {
+	if (ttyio < 0) /*EMPTY*/;
+#if	!MSDOS
+	else if (duptty[0]) loadtermio(ttyio, duptty[0], NULL);
+#endif
+	else {
 #if	MSDOS
-	raw2();
+		raw2();
 #else	/* !MSDOS */
 # ifdef	USESGTTY
-	raw2();
-	noecho2();
-	nonl2();
-	notabs();
+		raw2();
+		noecho2();
+		nonl2();
+		notabs();
+		if (dumbterm > 1) ttymode(ECHO | CRMOD, 0, 0, 0);
 # else	/* !USESGTTY */
-	if (isnl) ttymode(0, (ISIG|ICANON|IEXTEN) | (ECHO|ECHOE|ECHOK|ECHONL),
-		IGNBRK, (BRKINT|IXON) | ICRNL,
-		OPOST | ONLCR | TAB3, 0, 1, 0);
-	else ttymode(0, (ISIG|ICANON|IEXTEN) | (ECHO|ECHOE|ECHOK|ECHONL),
-		IGNBRK, (BRKINT|IXON) | ICRNL,
-		TAB3, OPOST | ONLCR, 1, 0);
+		if (isnl) ttymode(0, TIO_LCOOKED | ~TIO_LNOECHO,
+			IGNBRK, TIO_ICOOKED | ICRNL, TIO_ONL | TAB3, 0, 1, 0);
+		else ttymode(0, TIO_LCOOKED | ~TIO_LNOECHO,
+			IGNBRK, TIO_ICOOKED | ICRNL, TAB3, TIO_ONL, 1, 0);
+		if (dumbterm > 1)
+			ttymode(ECHO, 0, ICRNL, 0, TIO_ONL, 0, 1, 0);
 # endif	/* !USESGTTY */
 #endif	/* !MSDOS */
+	}
 	if (!dumbterm) {
 		putterm(T_KEYPAD);
 		tflush();
@@ -1117,30 +1126,33 @@ int stdiomode(VOID_A)
 	int isnl;
 
 	isnl = (isttyiomode) ? isttyiomode - 1 : 0;
+	if (ttyio < 0) /*EMPTY*/;
+#if	!MSDOS
+	else if (duptty[1]) loadtermio(ttyio, duptty[1], NULL);
+#endif
+	else {
 #if	MSDOS
-	cooked2();
+		cooked2();
 #else	/* !MSDOS */
 # ifdef	USESGTTY
-	cooked2();
-	echo2();
-	nl2();
-	tabs();
-	if (dumbterm > 2) ttymode(0, ECHO | CRMOD, 0, 0);
+		cooked2();
+		echo2();
+		nl2();
+		tabs();
+		if (dumbterm > 2) ttymode(0, ECHO | CRMOD, 0, 0);
 # else	/* !USESGTTY */
-	if (isnl) ttymode((ISIG|ICANON|IEXTEN) | (ECHO|ECHOE|ECHOCTL|ECHOKE),
-		PENDIN | ECHONL,
-		(BRKINT|IXON) | ICRNL, (IGNBRK|ISTRIP),
-		0, (OCRNL|ONOCR|ONLRET) | TAB3,
-		VAL_VMIN, VAL_VTIME);
-	else ttymode((ISIG|ICANON|IEXTEN) | (ECHO|ECHOE|ECHOCTL|ECHOKE),
-		PENDIN | ECHONL,
-		(BRKINT|IXON) | ICRNL, (IGNBRK|ISTRIP),
-		OPOST | ONLCR, (OCRNL|ONOCR|ONLRET) | TAB3,
-		VAL_VMIN, VAL_VTIME);
-	if (dumbterm > 2)
-		ttymode(0, ECHO, 0, ICRNL, 0, ONLCR, VAL_VMIN, VAL_VTIME);
+		if (isnl) ttymode(TIO_LCOOKED | TIO_LECHO, PENDIN | ECHONL,
+			TIO_ICOOKED | ICRNL, ~TIO_INOCOOKED,
+			0, ~TIO_ONONL | TAB3, VAL_VMIN, VAL_VTIME);
+		else ttymode(TIO_LCOOKED | TIO_LECHO, PENDIN | ECHONL,
+			TIO_ICOOKED | ICRNL, ~TIO_INOCOOKED,
+			TIO_ONL, ~TIO_ONONL | TAB3, VAL_VMIN, VAL_VTIME);
+		if (dumbterm > 2)
+			ttymode(0, ECHO,
+				0, ICRNL, 0, ONLCR, VAL_VMIN, VAL_VTIME);
 # endif	/* !USESGTTY */
 #endif	/* !MSDOS */
+	}
 	if (!dumbterm) {
 		putterm(T_NOKEYPAD);
 		tflush();
@@ -1336,7 +1348,7 @@ int *yp, *xp;
 int getxy(xp, yp)
 int *xp, *yp;
 {
-	char *format, buf[sizeof(SIZEFMT) + 6];
+	char *format, buf[sizeof(SIZEFMT) + (MAXLONGWIDTH - 2) * 2];
 	int i, j, tmp, count, *val[2];
 
 	format = SIZEFMT;
@@ -1554,8 +1566,18 @@ char *s;
 	char *p;
 
 	p = strbuf;
+#  ifdef	DEBUG
+	_mtrace_file = "tgetstr(start)";
 	s = tgetstr(s, &p);
-# endif
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "tgetstr(end)";
+		malloc(0);	/* dummy alloc */
+	}
+#  else
+	s = tgetstr(s, &p);
+#  endif
+# endif	/* !USETERMINFO */
 	if (s) {
 		if (*term) free(*term);
 		*term = tstrdup(s);
@@ -1570,14 +1592,35 @@ char **term, *str1, *str2;
 # ifdef	USETERMINFO
 	if (str1) str1 = tstrdup(str1);
 	else str1 = tparamstr(str2, 1, 1);
-# else
+# else	/* !USETERMINFO */
 	char *p, strbuf[TERMCAPSIZE];
 
 	p = strbuf;
 
+#  ifdef	DEBUG
+	_mtrace_file = "tgetstr(start)";
+	str1 = tgetstr(str1, &p);
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "tgetstr(end)";
+		malloc(0);	/* dummy alloc */
+	}
+	if (str1) str1 = tstrdup(str1);
+	else {
+		_mtrace_file = "tgetstr(start)";
+		str2 = tgetstr(str2, &p);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "tgetstr(end)";
+			malloc(0);	/* dummy alloc */
+		}
+		str1 = tparamstr(str2, 1, 1);
+	}
+#  else
 	if ((str1 = tgetstr(str1, &p))) str1 = tstrdup(str1);
 	else str1 = tparamstr(tgetstr(str2, &p), 1, 1);
-# endif
+#  endif
+# endif	/* !USETERMINFO */
 	if (str1) {
 		if (*term) free(*term);
 		*term = str1;
@@ -2574,13 +2617,12 @@ long usec;
 {
 #if	defined (NOTUSEBIOS) || !defined (DJGPP) || (DJGPP < 2)
 	union REGS reg;
-#else
+#else	/* !NOTUSEBIOS && DJGPP && DJGPP >= 2 */
 # ifndef	PC98
-	fd_set readfds;
 	struct timeval tv;
-	int n, fd;
+	int n;
 # endif
-#endif
+#endif	/* !NOTUSEBIOS && DJGPP && DJGPP >= 2 */
 
 	if (ungetnum > 0) return(1);
 #ifdef	NOTUSEBIOS
@@ -2601,17 +2643,9 @@ long usec;
 	return(reg.h.bh != 0);
 # else	/* !PC98 */
 #  if	defined (DJGPP) && (DJGPP >= 2)
-	fd = (ttyio < MAXFDSET || (n = safe_dup(ttyio)) < 0) ? ttyio : n;
 	tv.tv_sec = (time_t)usec / (time_t)1000000;
 	tv.tv_usec = (time_t)usec % (time_t)1000000;
-	FD_ZERO(&readfds);
-	FD_SET(fd, &readfds);
-
-	do {
-		n = select(fd + 1, &readfds, NULL, NULL, &tv);
-	} while (n < 0 && errno == EINTR);
-	if (fd != ttyio) close(fd);
-	if (n < 0) err2("select()");
+	if ((n = readselect(1, &ttyio, NULL, &tv)) < 0) err2("select()");
 
 	return(n);
 #  else	/* !DJGPP || DJGPP < 2 */
@@ -2734,7 +2768,7 @@ int sig;
 			raise(sig);
 		}
 # endif
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 	} while (!i);
 
 #else	/* !DJGPP || NOTUSEBIOS || PC98 */
@@ -2747,7 +2781,7 @@ int sig;
 # endif
 			memcpy((char *)tbuf1, (char *)tbuf2, sizeof(tbuf1));
 		}
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 	}
 #endif	/* !DJGPP || NOTUSEBIOS || PC98 */
 	if ((ch = getch2()) == EOF) return(K_NOKEY);
@@ -2783,7 +2817,11 @@ int c;
 {
 	if (ungetnum >= (int)sizeof(ungetbuf) / sizeof(u_char) - 1)
 		return(EOF);
-	ungetbuf[ungetnum++] = c;
+	if (ungetnum)
+		memmove(&(ungetbuf[1]), &(ungetbuf[0]),
+			ungetnum * sizeof(u_char));
+	ungetbuf[0] = c;
+	ungetnum++;
 
 	return(c);
 }
@@ -2881,21 +2919,12 @@ long usec;
 # ifdef	NOSELECT
 	return((usec) ? 1 : 0);
 # else
-	fd_set readfds;
 	struct timeval tv;
-	int n, fd;
+	int n;
 
-	fd = (ttyio < MAXFDSET || (n = safe_dup(ttyio)) < 0) ? ttyio : n;
 	tv.tv_sec = (time_t)usec / (time_t)1000000;
 	tv.tv_usec = (time_t)usec % (time_t)1000000;
-	FD_ZERO(&readfds);
-	FD_SET(fd, &readfds);
-
-	do {
-		n = select(fd + 1, &readfds, NULL, NULL, &tv);
-	} while (n < 0 && errno == EINTR);
-	if (fd != ttyio) close(fd);
-	if (n < 0) err2("select()");
+	if ((n = readselect(1, &ttyio, NULL, &tv)) < 0) err2("select()");
 
 	return(n);
 # endif
@@ -2934,7 +2963,7 @@ int sig;
 			ttyiomode((isttyiomode) ? isttyiomode - 1 : 0);
 			suspended = 0;
 		}
-		if (keywaitfunc && (*keywaitfunc)() < 0) return(-1);
+		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
 # ifndef	TIOCSTI
 		if (ungetnum > 0) return((int)ungetbuf[--ungetnum]);
 # endif
@@ -2965,10 +2994,12 @@ int sig;
 			return(mkmetakey(ch));
 		for (j = 0; j < p -> num; j++)
 			if (key == keyseq[p -> next[j].key].str[0]) break;
-		if (j >= p -> num) return(key);
+		if (j >= p -> num) {
+			ungetch2(ch);
+			return(key);
+		}
 		p = &(p -> next[j]);
-		if (keyseq[p -> key].len == 1)
-			return(keyseq[p -> key].code);
+		if (keyseq[p -> key].len == 1) return(keyseq[p -> key].code);
 	}
 	else {
 		for (j = 0; j < p -> num; j++)
@@ -2984,14 +3015,19 @@ int sig;
 	for (i = 1; p && p -> next; i++) {
 		for (j = 0; j < p -> num; j++)
 			if (ch == keyseq[p -> next[j].key].str[i]) break;
-		if (j >= p -> num) return(key);
+		if (j >= p -> num) break;
 		p = &(p -> next[j]);
 		if (keyseq[p -> key].len == i + 1)
 			return(keyseq[p -> key].code);
 		if (!kbhit2(WAITKEYPAD * 1000L) || (ch = getch2()) == EOF)
-			return(key);
+			break;
 	}
 
+	for (j = 1; j < i; j++) {
+		if (j >= keyseq[p -> key].len) break;
+		ungetch2(keyseq[p -> key].str[j]);
+	}
+	ungetch2(ch);
 	return(key);
 }
 
@@ -3020,7 +3056,11 @@ int c;
 # else
 	if (ungetnum >= (int)sizeof(ungetbuf) / sizeof(u_char) - 1)
 		return(EOF);
-	ungetbuf[ungetnum++] = c;
+	if (ungetnum)
+		memmove(&(ungetbuf[1]), &(ungetbuf[0]),
+			ungetnum * sizeof(u_char));
+	ungetbuf[0] = c;
+	ungetnum++;
 # endif
 
 	return(c);
@@ -3096,7 +3136,8 @@ int xmax, ymax;
 
 	if (dumbterm) /*EMPTY*/;
 	else if (usegetcursor || x < 0 || y < 0) {
-		VOID_C setscroll(-1, -1);
+		if (usegetcursor) VOID_C setscroll(-1, -1);
+		else VOID_C setscroll(0, 998);
 		if (maxlocate(&ty, &tx) >= 0 && (tx > x || ty > y)) {
 			if (tx > x) x = tx;
 			if (ty > y) y = ty;
@@ -3115,7 +3156,7 @@ int xmax, ymax;
 	if (n_line <= 0 || (ymax > 0 && n_line < ymax))
 		return("Line size too small");
 
-	if (xmax > 0 && ymax > 0) VOID_C setscroll(-1, -1);
+	if (xmax > 0 && ymax > 0) VOID_C setscroll(0, n_line - 1);
 
 	return(NULL);
 }
@@ -3171,10 +3212,10 @@ va_dcl
 	int n;
 
 	VA_START(args, fmt);
-
 	n = vasprintf2(&buf, fmt, args);
 	va_end(args);
 	if (n < 0) err2("malloc()");
+
 	cputs2(buf);
 	free(buf);
 
