@@ -37,8 +37,9 @@ typedef struct fd_set {
 # define	FD_SET(n, p)	(((p) -> fds_bits[0]) |= ((u_int)1 << (n)))
 #endif	/* !FD_SET */
 
-extern int hideclock;
 extern int internal_status;
+extern int hideclock;
+extern int fdmode;
 
 int ptymode = 0;
 char *ptyterm = NULL;
@@ -238,17 +239,25 @@ static int NEAR genbackend(VOID_A)
 	return(0);
 }
 
-VOID syncptyout(VOID_A)
+VOID syncptyout(fd, cmd)
+int fd, cmd;
 {
 	char *tty;
+	u_char uc;
 
 	if (parentfd < 0) return;
 
 	savetermio(ttyio, &tty, NULL);
 	keyflush();
 	noecho2();
-	tputs2("\033[99n", 1);
-	tflush();
+	if (fd < 0) {
+		tputs2("\033[99n", 1);
+		tflush();
+	}
+	else {
+		uc = cmd;
+		sendbuf(fd, &uc, sizeof(uc));
+	}
 	VOID_C getch2();
 	keyflush();
 	loadtermio(ttyio, tty, NULL);
@@ -448,6 +457,7 @@ va_dcl
 #endif
 
 	fd = parentfd;
+	syncptyout(fd, TE_LOCKFRONT);
 	uc = cmd;
 	sendbuf(fd, &uc, sizeof(uc));
 
@@ -600,10 +610,18 @@ va_dcl
 			sendstring(fd, devp -> name);
 			break;
 #endif	/* _USEDOSEMU */
+		case TE_SAVETTYIO:
+			n = va_arg(args, int);
+			val = (n >= 0 && duptty[n]) ? TIO_BUFSIZ : 0;
+			sendbuf(fd, &n, sizeof(n));
+			sendbuf(fd, &val, sizeof(val));
+			if (n >= 0 && duptty[n]) sendbuf(fd, duptty[n], val);
+			break;
 		default:
 			break;
 	}
 	va_end(args);
+	syncptyout(fd, TE_UNLOCKFRONT);
 }
 
 static VOID NEAR awakechild(command, arg, flags)
@@ -705,10 +723,17 @@ char *command, *arg;
 int flags;
 {
 	p_id_t pid;
-	char *tty, *ws, path[MAXPATHLEN];
+	char *tty, *ws, path[MAXPATHLEN], buf[TIO_BUFSIZ + TIO_WINSIZ];
 	u_char uc;
 	int i, n, fd, fds[2];
 
+#ifndef	_NOORIGSHELL
+	if (isshptymode()) {
+		if (!(flags & F_DOSYSTEM))
+			return(callmacro(command, arg, flags));
+	}
+	else
+#endif
 	if (!ptymode) return(callmacro(command, arg, flags));
 
 	if (ptylist[win].pid && emufd >= 0) {
@@ -716,6 +741,18 @@ int flags;
 		return(frontend());
 	}
 
+	savetermio(ttyio, &tty, &ws);
+	if (tty) {
+		memcpy(buf, tty, TIO_BUFSIZ);
+		free(tty);
+		tty = buf;
+	}
+	if (ws) {
+		memcpy(&(buf[TIO_BUFSIZ]), ws, TIO_WINSIZ);
+		free(ws);
+		ws = &(buf[TIO_BUFSIZ]);
+	}
+
 	if (genbackend() < 0 || pipe(fds) < 0)
 		return(callmacro(command, arg, flags));
 	VOID_C fcntl(fds[0], F_SETFL, O_NONBLOCK);
@@ -724,9 +761,14 @@ int flags;
 	if (ptytmpfile) fd = -1;
 	else if ((fd = mktmpfile(path)) >= 0) {
 		ptytmpfile = strdup2(path);
-		Xclose(fd);
+		VOID_C Xclose(fd);
 	}
 
+#ifndef	_NOKANJICONV
+	changeinkcode();
+	changeoutkcode();
+#endif
+
 	n = sigvecset(0);
 	if ((pid = Xfork()) < (p_id_t)0) {
 		if (fd >= 0 && ptytmpfile) {
@@ -744,10 +786,11 @@ int flags;
 #ifndef	_NOORIGSHELL
 		mypid = getpid();
 #endif
-		savetermio(ttyio, &tty, &ws);
 		if (Xlogin_tty(ptylist[win].path, tty, ws) < 0) _exit(1);
-		if (tty) free(tty);
-		if (ws) free(ws);
+#ifndef	_NOORIGSHELL
+		if (isshptymode()) /*EMPTY*/;
+		else
+#endif
 		n_line = FILEPERROW;
 		VOID_C setwsize(STDIN_FILENO, n_column, n_line);
 
@@ -781,11 +824,9 @@ int flags;
 		sendbuf(parentfd, &uc, sizeof(uc));
 
 		for (;;) {
-			syncptyout();
+			syncptyout(-1, -1);
 			setlinecol();
-			n = callmacro(command, arg, flags);
-			n = (n < 0) ? 1 :
-				((internal_status < -1) ? 4 : internal_status);
+			n = evalstatus(callmacro(command, arg, flags));
 #if	defined (_NOORIGSHELL) || defined (NOJOB)
 			break;
 #else	/* !_NOORIGSHELL && !NOJOB */
