@@ -10,6 +10,9 @@ typedef struct _ptyinfo_t {
 	int fd;
 	int pipe;
 	int status;
+#ifndef	_NOKANJICONV
+	char incode, outcode;
+#endif
 } ptyinfo_t;
 
 extern ptyinfo_t ptylist[];
@@ -27,7 +30,11 @@ extern ptyinfo_t ptylist[];
 #define	TE_CHANGEWSIZE		(K_MAX + 11)
 #define	TE_INSERTWIN		(K_MAX + 12)
 #define	TE_DELETEWIN		(K_MAX + 13)
-#define	TE_CHANGEKCODE		(K_MAX + 14)
+#define	TE_LOCKBACK		(K_MAX + 14)
+#define	TE_UNLOCKBACK		(K_MAX + 15)
+#define	TE_CHANGEKCODE		(K_MAX + 16)
+#define	TE_CHANGEINKCODE	(K_MAX + 17)
+#define	TE_CHANGEOUTKCODE	(K_MAX + 18)
 #define	TE_AWAKECHILD		(K_MAX + 99)
 
 #define	TE_SETVAR		1
@@ -54,4 +61,8 @@ extern ptyinfo_t ptylist[];
 #define	TE_DELETEARCH		22
 #define	TE_INSERTDRV		23
 #define	TE_DELETEDRV		24
+#define	TE_LOCKFRONT		25
+#define	TE_UNLOCKFRONT		26
+#define	TE_SAVETTYIO		27
+#define	TE_INTERNAL		28
 #define	TE_CHANGESTATUS		99
